import sys

import numpy as np
import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.buttons as buttons

import logging
log = logging.getLogger(__name__)


class TileCategory(object):
    def __init__(self, name):
        self.name = name

    def get_height(self, parent):
        w, h = parent.GetTextExtent(self.name)
        return h + 5

    def draw(self, font, dc, rect):
        dc.DrawLabel(self.name, rect, wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)


class Tile(object):
    def __init__(self, tile_num, keystroke):
        self.tile_num = tile_num
        self.keystroke = keystroke

    def get_bytes(self):
        return (self.tile_num, )  # has to be tuple so it's hashable

    def get_height(self, parent):
        if parent.editor:  # Might get called before an editor is set
            return parent.editor.machine.antic_font.get_height(parent.zoom)
        return 10

    def draw(self, parent, dc, rect):
        if parent.editor:  # Might get called before an editor is set
            bmp = parent.editor.machine.antic_font.get_image(self.tile_num, parent.zoom)
            dc.DrawBitmap(bmp, rect.x+10, rect.y)


class TileListBox(wx.VListBox):
    # This method must be overridden.  When called it should draw the
    # n'th item on the dc within the rect.  How it is drawn, and what
    # is drawn is entirely up to you.
    def OnDrawItem(self, dc, rect, n):
        if self.GetSelection() == n:
            c = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            c = self.GetForegroundColour()
        dc.SetFont(self.GetFont())
        dc.SetTextForeground(c)
        item = self.GetParent().items[n]
        item.draw(self.GetParent(), dc, rect)

    # This method must be overridden.  It should return the height
    # required to draw the n'th item.
    def OnMeasureItem(self, n):
        item = self.GetParent().items[n]
        h = item.get_height(self.GetParent())
        return h


class TileListControl(wx.Panel):
    """
    View for displaying categories of tiles to paint the map
    """

    def __init__(self, parent, task, command=None, **kwargs):
        wx.Panel.__init__(self, parent, -1, **kwargs)
        self.task = task
        self.editor = None
        self.command_cls = command

        self.tile_list = TileListBox(self, -1, style=wx.VSCROLL)
        self.tile_list.Bind(wx.EVT_LEFT_DOWN, self.on_tile_clicked)
        self.tile_list.Bind(wx.EVT_LISTBOX, self.on_tile)
        self.cat = wx.Choice(self, -1)
        self.cat.Bind(wx.EVT_CHOICE, self.on_category)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.cat, 0, wx.EXPAND, 0)
        sizer.Add(self.tile_list, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Fit()

        self.pattern_to_index = {}
        self.parse_tile_map([("test", np.arange(0,10, dtype=np.uint8))])
        self.current_tile = -1
        self.setup_tiles()
        self.zoom = 2

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.parse_tile_map(editor.antic_tile_map)
            self.setup_tiles()

    def parse_tile_map(self, tile_map):
        self.tile_map = tile_map
        self.categories = []
        self.items = []
        for label, tiles in tile_map:
            item = TileCategory(label)
            self.categories.append((item, len(self.items)))
            self.items.append(item)
            for i in np.arange(np.alen(tiles)):
                item = Tile(tiles[i:i+1], "")
                self.items.append(item)
                self.pattern_to_index[item.get_bytes] = i

    def setup_tiles(self):
        cats = [c[0].name for c in self.categories]
        self.cat.Set(cats)
        self.cat.SetSelection(0)
        self.tile_list.SetItemCount(len(self.items))
        self.tile_list.SetSelection(self.current_tile)

    def on_category(self, event):
        cat = event.GetSelection()
        item, index = self.categories[cat]
        self.tile_list.ScrollToRow(index)

    def on_tile_clicked(self, event):
        p = event.GetPosition()
        index = self.tile_list.HitTest(p)
        e = self.editor
        if e is not None:
            self.current_tile = index
            wx.CallAfter(e.set_current_draw_pattern, self.items[index].get_bytes(), self)

    def show_pattern(self, pattern):
        index = self.index_from_pattern.get(pattern, None)
        if index is not None:
            self.tile_list.SetSelection(index)
            self.tile_list.ScrollToRow(index)

    def clear_tile_selection(self):
        self.current_tile = -1
        self.tile_list.SetSelection(-1)

    def on_tile(self, event):
        index = event.GetInt()
        self.change_tile(self.items[index])

    def change_tile(self, tile):
        e = self.editor
        if e is None:
            return
        cmd_cls = self.command_cls
        if cmd_cls is None:
            return
        if e.can_copy:
            index = e.anchor_start_index
        else:
            index = e.cursor_index
        value = tile.get_bytes()
        cmd = cmd_cls(e.segment, index, index+len(value), value, True)
        e.process_command(cmd)


class TileButton(buttons.GenBitmapToggleButton):
    labelDelta = 0
    label_border = 4
    faceDnClr = wx.Colour(100, 200, 230)

    def DoGetBestSize(self):
        """
        Overridden base class virtual.  Determines the best size of the
        button based on the label and bezel size.
        """
        w, h, useMin = self._GetLabelSize()
        width = w + self.label_border + 2 * self.bezelWidth + 4 * int(self.useFocusInd)
        height = h + self.label_border + 2 * self.bezelWidth + 4 * int(self.useFocusInd)
        return (width, height)

    def InitColours(self):
        pass

    def DrawBezel(self, dc, x1, y1, x2, y2):
        pass

    @classmethod
    def set_colors(cls, editor):
        cls.faceDnClr = wx.Colour(*editor.machine.highlight_color)


class TileWrapControl(wx.Panel):
    """
    View for displaying categories of tiles to paint the map
    """

    def __init__(self, parent, task, command=None, **kwargs):
        wx.Panel.__init__(self, parent, -1, **kwargs)
        self.task = task
        self.editor = None
        self.command_cls = command

        self.cat = wx.Choice(self, -1)
        self.cat.Bind(wx.EVT_CHOICE, self.on_category)
        self.panel = scrolled.ScrolledPanel(self, -1, style=wx.VSCROLL)
        self.panel.Bind(wx.EVT_SIZE, self.panel_size)
        psiz = wx.BoxSizer(wx.VERTICAL)
        self.panel.SetSizer(psiz)
        self.panel.ShowScrollbars(wx.SHOW_SB_NEVER, wx.SHOW_SB_ALWAYS)
        self.panel.SetupScrolling(scroll_x=False)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.cat, 0, wx.EXPAND, 0)
        sizer.Add(self.panel, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Fit()

        self.pattern_to_item = {}
        self.parse_tile_map(self.panel, [("test", np.arange(0,10, dtype=np.uint8))])
        self.current_tile = None
        self.setup_tiles()
        self.zoom = 2

    def panel_size(self, evt):
        size = self.panel.GetSize()
        vsize = self.panel.GetVirtualSize()
        self.panel.SetVirtualSize((size[0] - 20, vsize[1]))

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.parse_tile_map(self.panel, editor.antic_tile_map)
            self.setup_tiles()
            TileButton.set_colors(editor)

    def parse_tile_map(self, panel, tile_map):
        sizer = panel.GetSizer()
        sizer.DeleteWindows()
        self.tile_map = tile_map
        self.categories = []
        self.items = []
        self.pattern_to_item = {}
        for items in tile_map:
            label = items[0]
            t = wx.StaticText(panel, -1, label)
            sizer.Add(t, 0, wx.EXPAND, 0)
            self.categories.append(t)
            w = wx.WrapSizer()
            for tiles in items[1:]:
                for i in np.arange(np.alen(tiles)):
                    if self.editor:
                        data = tiles[i:i+1]
                        bmp = self.editor.machine.antic_font.get_image(data[0], self.zoom)
                        btn = TileButton(panel, -1, bmp, style=wx.BORDER_NONE|wx.BU_EXACTFIT)
                        btn.tile_data = data
                        btn.Bind(wx.EVT_BUTTON, self.on_tile_clicked)
                        w.Add(btn, 0, wx.ALL, 0)
                        self.items.append(btn)
                        self.pattern_to_item[tuple(data)] = btn
            sizer.Add(w, 0, wx.EXPAND, 0)
        self.Layout()

    def setup_tiles(self):
        cats = [str(c.GetLabelText()) for c in self.categories]
        self.cat.Set(cats)
        self.cat.SetSelection(0)

    def scroll_to_control(self, ctrl):
        sppu_x, sppu_y = self.panel.GetScrollPixelsPerUnit()
        vs_x, vs_y = self.panel.GetViewStart()
        cr = ctrl.GetRect()
        self.panel.Scroll(0, vs_y + (cr.y / sppu_y))

    def on_category(self, event):
        cat = event.GetSelection()
        label = self.categories[cat]
        self.scroll_to_control(label)

    def on_tile_clicked(self, event):
        btn = event.GetEventObject()
        e = self.editor
        if e is not None:
            self.current_tile = btn
            e.set_current_draw_pattern(btn.tile_data, self)
            self.clear_toggle_except(btn)

    def show_pattern(self, pattern):
        log.debug("tilelist showing pattern %s" % str(pattern))
        btn = self.pattern_to_item.get(pattern, None)
        self.clear_toggle_except(btn)
        if btn is not None:
            self.scroll_to_control(btn)

    def clear_tile_selection(self):
        self.current_tile = None
        self.clear_toggle_except()

    def clear_toggle_except(self, btn=None):
        for b in self.items:
            if b != btn:
                b.SetToggle(False)
        if btn is not None:
            btn.SetToggle(True)

    def on_tile(self, event):
        index = event.GetInt()
        self.change_tile(self.items[index])

    def change_tile(self, tile):
        e = self.editor
        if e is None:
            return
        cmd_cls = self.command_cls
        if cmd_cls is None:
            return
        if e.can_copy:
            index = e.anchor_start_index
        else:
            index = e.cursor_index
        value = tile.get_bytes()
        cmd = cmd_cls(e.segment, index, index+len(value), value, True)
        e.process_command(cmd)


if __name__ == '__main__':
    import sys
    sys.path[0:0] = [".."]
    print sys.path
    import fonts
    import colors

    class Wrapper(object):
        def __init__(self, **kwargs):
            for k, v in kwargs.iteritems():
                setattr(self, k, v)

        def set_current_draw_pattern(self, *args, **kwargs):
            pass

    class MyFrame(wx.Frame):
        def __init__(self, parent, id, title):
            wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition, wx.DefaultSize)
            tile_map = [
            ("road", [0x70]),
            ("trees", range(0x80, 0x96), range(0x01, 0x16),),
            ("buildings", range(0x96, 0x9f), range(0x16, 0x1f), range(0x41, 0x51), range(0x5d, 0x60),),
            ("people", range(0xf1, 0xf4), range(0x71, 0x74)),
            ("water", range(0x2e, 0x41),),
            ("bridges", range(0x69, 0x6d),),
            ("vehicles", range(0x51, 0x59),),
            ("airport", range(0x60, 0x68), [0x5f], range(0x59, 0x5d), range(0xd9, 0xdd)),
            ("golf", range(0xa9, 0xae),),
            ("other", [0x20, 0x25, 0x26, ]),
            ("special", range(0x21, 0x25), range(0x74, 0x76),),
                ]
            color_converter = colors.gtia_ntsc_to_rgb
            highlight_color = (100, 200, 230)
            unfocused_cursor_color = (128, 128, 128)
            background_color = (255, 255, 255)
            match_background_color = (255, 255, 180)
            comment_background_color = (255, 180, 200)
            antic_font = fonts.AnticFont(fonts.A8DefaultFont, 4, colors.powerup_colors(), highlight_color, match_background_color, comment_background_color, color_converter)
            editor = Wrapper(antic_font=antic_font, antic_tile_map=tile_map, highlight_color=highlight_color)
            task = Wrapper(active_editor=editor)
            panel = TileWrapControl(self, task)
            panel.recalc_view()

    class MyApp(wx.App):
        def OnInit(self):
            frame = MyFrame(None, -1, 'test')
            frame.Show(True)
            frame.Center()
            return True

    app = MyApp(0)
    app.MainLoop()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" The wx specific implementation of a menu bar manager.
"""

# Major package imports.
import wx

# Local imports.
from pyface.action.action_manager import ActionManager


class MenuBarManager(ActionManager):
    """ A menu bar manager realizes itself in errr, a menu bar control. """

    ###########################################################################
    # 'MenuBarManager' interface.
    ###########################################################################

    def create_menu_bar(self, parent, controller=None):
        """ Creates a menu bar representation of the manager. """

        # If a controller is required it can either be set as a trait on the
        # menu bar manager (the trait is part of the 'ActionManager' API), or
        # passed in here (if one is passed in here it takes precedence over the
        # trait).
        if controller is None:
            controller = self.controller

        menu_bar = wx.MenuBar()

        # Every item in every group must be a menu manager.
        for group in self.groups:
            for item in group.items:
                menu = item.create_menu(parent, controller)
                menu_bar.Append(menu, item.name)

        return menu_bar

#### EOF ######################################################################



"""
Miscellaneous util functions.

"""

from __future__ import (absolute_import, division, print_function)

import numpy as np


epsilon = 1.e-5


def point_in_tri(face_points, point, return_weights=False):
    """
    Calculates whether point is internal/external
    to element by comparing summed area of sub triangles with area of triangle
    element.

    """
    sub_tri_areas = np.zeros(3)
    sub_tri_areas[0] = _signed_area_tri(np.vstack((face_points[(0, 1), :],
                                                   point)))
    sub_tri_areas[1] = _signed_area_tri(np.vstack((face_points[(1, 2), :],
                                                   point)))
    sub_tri_areas[2] = _signed_area_tri(np.vstack((face_points[(0, 2), :],
                                                   point)))
    tri_area = _signed_area_tri(face_points)

    if abs(np.abs(sub_tri_areas).sum()-tri_area)/tri_area <= epsilon:
        if return_weights:
            raise NotImplementedError
            # weights = sub_tri_areas/tri_area
            # weights[1] = max(0., min(1., weights[1]))
            # weights[2] = max(0., min(1., weights[2]))
            # if (weights[0]+weights[1]>1):
            #     weights[2] = 0
            #     weights[1] = 1-weights[0]
            # else:
            #     weights[2] = 1-weights[0]-weights[1]
            #
            # return weights
        else:
            return True
    return False


def _signed_area_tri(points):
    """
    points : the coordinates of the triangle vertices -- (3x2) float array
    returns signed area of the triangle.

    """

    x1, y1 = points[0]
    x2, y2 = points[1]
    x3, y3 = points[2]

    return(((x1-x3)*(y2-y3)-(x2-x3)*(y1-y3))/2)


must_have = ['dtype', 'shape', 'ndim','__len__', '__getitem__', '__getattribute__']
def isarraylike(obj):
    """
    tests if obj acts enough like an array to be used in pyugrid.

    This should catch netCDF4 variables, etc.

    Note: these won't check if the attributes required actually work right.
    """
    for attr in must_have:
        if not hasattr(obj, attr):

            return False

    return True

def asarraylike(obj):
    """
    If it satisfies the requirements of pyugrid the object is returned as is. If not, then numpy's
    array() will be called on it.

    :param obj: The object to check if it's like an array

    """

    return obj if isarraylike(obj) else np.array(obj)





"""
fs.expose.importhook
====================

Expose an FS object to the python import machinery, via a PEP-302 loader.

This module allows you to import python modules from an arbitrary FS object,
by placing FS urls on sys.path and/or inserting objects into sys.meta_path.

The main class in this module is FSImportHook, which is a PEP-302-compliant
module finder and loader.  If you place an instance of FSImportHook on
sys.meta_path, you will be able to import modules from the exposed filesystem::

    >>> from fs.memoryfs import MemoryFS
    >>> m = MemoryFS()
    >>> m.setcontents("helloworld.py","print 'hello world!'")
    >>>
    >>> import sys
    >>> from fs.expose.importhook import FSImportHook
    >>> sys.meta_path.append(FSImportHook(m))
    >>> import helloworld
    hello world!

It is also possible to install FSImportHook as an import path handler. This
allows you to place filesystem URLs on sys.path and have them automagically
opened for importing.  This example would allow modules to be imported from
an SFTP server::

    >>> from fs.expose.importhook import FSImportHook
    >>> FSImportHook.install()
    >>> sys.path.append("sftp://some.remote.machine/mypath/")

"""

import os
import sys
import imp
import marshal

from fs.base import FS
from fs.opener import fsopendir, OpenerError
from fs.errors import *
from fs.path import *

from six import b

class FSImportHook(object):
    """PEP-302-compliant module finder and loader for FS objects.

    FSImportHook is a module finder and loader that takes its data from an
    arbitrary FS object.  The FS must have .py or .pyc files stored in the
    standard module structure.

    For easy use with sys.path, FSImportHook will also accept a filesystem
    URL, which is automatically opened using fs.opener.
    """

    _VALID_MODULE_TYPES = set((imp.PY_SOURCE,imp.PY_COMPILED))

    def __init__(self,fs_or_url):
        #  If given a string, try to open it as an FS url.
        #  Don't open things on the local filesystem though.
        if isinstance(fs_or_url,basestring):
            if ":/" not in fs_or_url:
                raise ImportError
            try:
                self.fs = fsopendir(fs_or_url)
            except OpenerError:
                raise ImportError
            except (CreateFailedError,ResourceNotFoundError,):
                raise ImportError
            self.path = fs_or_url
        #  Otherwise, it must be an FS object of some sort.
        else:
            if not isinstance(fs_or_url,FS):
                raise ImportError
            self.fs = fs_or_url
            self.path = None

    @classmethod
    def install(cls):
        """Install this class into the import machinery.

        This classmethod installs the custom FSImportHook class into the
        import machinery of the running process, if it is not already
        installed.
        """
        for imp in enumerate(sys.path_hooks):
            try:
                if issubclass(cls,imp):
                    break
            except TypeError:
                pass
        else:
            sys.path_hooks.append(cls)
            sys.path_importer_cache.clear()

    @classmethod
    def uninstall(cls):
        """Uninstall this class from the import machinery.

        This classmethod uninstalls the custom FSImportHook class from the
        import machinery of the running process.
        """
        to_rem = []
        for i,imp in enumerate(sys.path_hooks):
            try:
                if issubclass(cls,imp):
                    to_rem.append(imp)
                    break
            except TypeError:
                pass
        for imp in to_rem:
            sys.path_hooks.remove(imp)
            sys.path_importer_cache.clear()

    def find_module(self,fullname,path=None):
        """Find the FS loader for the given module.

        This object is always its own loader, so this really just checks
        whether it's a valid module on the exposed filesystem.
        """
        try:
            self._get_module_info(fullname)
        except ImportError:
            return None
        else:
            return self

    def _get_module_info(self,fullname):
        """Get basic information about the given module.

        If the specified module exists, this method returns a tuple giving
        its filepath, file type and whether it's a package.  Otherwise,
        it raise ImportError.
        """
        prefix = fullname.replace(".","/")
        #  Is it a regular module?
        (path,type) = self._find_module_file(prefix)
        if path is not None:
            return (path,type,False)
        #  Is it a package?
        prefix = pathjoin(prefix,"__init__")
        (path,type) = self._find_module_file(prefix)
        if path is not None:
            return (path,type,True)
        #  No, it's nothing
        raise ImportError(fullname)

    def _find_module_file(self,prefix):
        """Find a module file from the given path prefix.

        This method iterates over the possible module suffixes, checking each
        in turn and returning the first match found.  It returns a two-tuple
        (path,type) or (None,None) if there's no module.
        """
        for (suffix,mode,type) in imp.get_suffixes():
            if type in self._VALID_MODULE_TYPES:
                path = prefix + suffix
                if self.fs.isfile(path):
                    return (path,type)
        return (None,None)

    def load_module(self,fullname):
        """Load the specified module.

        This method locates the file for the specified module, loads and
        executes it and returns the created module object.
        """
        #  Reuse an existing module if present.
        try:
            return sys.modules[fullname]
        except KeyError:
            pass
        #  Try to create from source or bytecode.
        info = self._get_module_info(fullname)
        code = self.get_code(fullname,info)
        if code is None:
            raise ImportError(fullname)
        mod = imp.new_module(fullname)
        mod.__file__ = "<loading>"
        mod.__loader__ = self
        sys.modules[fullname] = mod
        try:
            exec code in mod.__dict__
            mod.__file__ = self.get_filename(fullname,info)
            if self.is_package(fullname,info):
                if self.path is None:
                    mod.__path__ = []
                else:
                    mod.__path__ = [self.path]
            return mod
        except Exception:
            sys.modules.pop(fullname,None)
            raise

    def is_package(self,fullname,info=None):
        """Check whether the specified module is a package."""
        if info is None:
            info = self._get_module_info(fullname)
        (path,type,ispkg) = info
        return ispkg

    def get_code(self,fullname,info=None):
        """Get the bytecode for the specified module."""
        if info is None:
            info = self._get_module_info(fullname)
        (path,type,ispkg) = info
        code = self.fs.getcontents(path, 'rb')
        if type == imp.PY_SOURCE:
            code = code.replace(b("\r\n"),b("\n"))
            return compile(code,path,"exec")
        elif type == imp.PY_COMPILED:
            if code[:4] != imp.get_magic():
                return None
            return marshal.loads(code[8:])
        else:
            return None
        return code

    def get_source(self,fullname,info=None):
        """Get the sourcecode for the specified module, if present."""
        if info is None:
            info = self._get_module_info(fullname)
        (path,type,ispkg) = info
        if type != imp.PY_SOURCE:
            return None
        return self.fs.getcontents(path, 'rb').replace(b("\r\n"),b("\n"))

    def get_data(self,path):
        """Read the specified data file."""
        try:
            return self.fs.getcontents(path, 'rb')
        except FSError, e:
            raise IOError(str(e))

    def get_filename(self,fullname,info=None):
        """Get the __file__ attribute for the specified module."""
        if info is None:
            info = self._get_module_info(fullname)
        (path,type,ispkg) = info
        return path
 



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/18/2004
#
#------------------------------------------------------------------------------

""" Define the ViewElements class, which is used to define a (typically
    class-based) hierarchical name space of related ViewElement objects.

    Normally there is a ViewElements object associated with each Traits-based
    class, which contains all of the ViewElement objects associated with the
    class. The ViewElements object is also linked to the ViewElements objects
    of its associated class's parent classes.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import HasStrictTraits, List, Dict, Str, Int, Any, TraitError

from .view_element import ViewElement

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Trait for contents of a ViewElements object
content_trait = Dict(str, ViewElement)

#-------------------------------------------------------------------------
#  'ViewElements' class:
#-------------------------------------------------------------------------


class ViewElements(HasStrictTraits):
    """ Defines a hierarchical name space of related ViewElement objects.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Dictionary containing the named ViewElement items
    content = content_trait

    #-------------------------------------------------------------------------
    #  Finds a specified ViewElement within the specified (optional) search
    #  context:
    #-------------------------------------------------------------------------

    def find(self, name, stack=None):
        """ Finds a specified ViewElement within the specified (optional) search
            context.
        """
        # Assume search starts from the beginning the of the search order:
        i = 0

        # If a stack was specified, see if there is a matching entry in the
        # stack already:
        if stack is not None:
            for ssi in stack:
                if name == ssi.id:
                    # Match found, resume search at next ViewElements object
                    # in the search order:
                    i = ssi.context + 1
                    break

        # Search for a matching name starting at the specified ViewElements
        # object in the search order:
        for j, ves in enumerate(self._get_search_order()[i:]):
            result = ves.content.get(name)
            if result is not None:
                # Match found. If there is a stack, push matching name and
                # ViewElements context onto it:
                if stack is not None:
                    stack[0:0] = [SearchStackItem(id=name,
                                                  context=i + j)]

                # Return the ViewElement object that matched the name:
                return result

        # Indicate no match was found:
        return None

    #-------------------------------------------------------------------------
    #  Returns a sorted list of all names accessible from the ViewElements
    #  object that are of a specified (ViewElement) type:
    #-------------------------------------------------------------------------

    def filter_by(self, klass=None):
        """ Returns a sorted list of all names accessible from the ViewElements
            object that are of a specified (ViewElement) type.
        """
        if klass is None:
            from . import view
            klass = view.View
        result = []

        # Add each item in the search order which is of the right class and
        # which is not already in the result list:
        for ves in self._get_search_order():
            for name, ve in ves.content.items():
                if isinstance(ve, klass) and (name not in result):
                    result.append(name)

        # Sort the resulting list of names:
        result.sort()

        # Return the result:
        return result

    #-------------------------------------------------------------------------
    #  Handles the 'parents' list being updated:
    #-------------------------------------------------------------------------

    def _parents__changed(self):
        self._search_order = None

    def _parents_items_changed(self):
        self._search_order = None

    #-------------------------------------------------------------------------
    #  Returns the current search order (computing it if necessary):
    #-------------------------------------------------------------------------

    def _get_search_order(self):
        if self._search_order is None:
            self._search_order = self._mro()
        return self._search_order

    #-------------------------------------------------------------------------
    #  Compute the Python 'C3' algorithm used to determine a class's 'mro'
    #  and apply it to the 'parents' of the ViewElements to determine the
    #  correct search order:
    #-------------------------------------------------------------------------

    def _mro(self):
        return self._merge(
            [[self]] +
            [parent._get_search_order()[:] for parent in self.parents] +
            [self.parents[:]])

    def _merge(self, seqs):
        result = []
        while True:
            # Remove any empty sequences from the list:
            seqs = [seq for seq in seqs if len(seq) > 0]
            if len(seqs) == 0:
                return result

            # Find merge candidates among the sequence heads:
            for seq in seqs:
                candidate = seq[0]
                if len([s for s in seqs if candidate in s[1:]]) == 0:
                    break
            else:
                raise TraitError("Inconsistent ViewElements hierarchy")

            # Add the candidate to the result:
            result.append(candidate)

            # Then remove the candidate:
            for seq in seqs:
                if seq[0] == candidate:
                    del seq[0]

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of the ViewElements object:
    #-------------------------------------------------------------------------

    def __repr__(self):
        """ Returns a "pretty print" version of the ViewElements object.
        """
        return self.content.__repr__()

#-------------------------------------------------------------------------
#  Define forward reference traits:
#-------------------------------------------------------------------------

ViewElements.add_class_trait('parents', List(ViewElements))
ViewElements.add_class_trait('_search_order', Any)

#-------------------------------------------------------------------------
#  'SearchStackItem' class:
#-------------------------------------------------------------------------


class SearchStackItem(HasStrictTraits):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Name that was looked up
    id = Str

    # Index into the 'mro' list of ViewElements that the ID was found in
    context = Int



""" Base class defining common code for EnamlTaskPane and EnamlEditor. """

# Enthought library imports.
from traits.api import HasTraits, Instance


class EnamlPane(HasTraits):
    """ Base class defining common code for EnamlTaskPane and EnamlEditor. """

    ###########################################################################
    # 'EnamlPane' interface
    ###########################################################################

    #: The Enaml component defining the contents of the TaskPane.
    component = Instance('enaml.widgets.toolkit_object.ToolkitObject')

    def create_component(self):
        """ Return an Enaml component defining the contents of the pane.

        Returns
        -------
        component : ToolkitObject
        """
        raise NotImplementedError

    ###########################################################################
    # 'TaskPane'/'Editor' interface
    ###########################################################################

    def create(self, parent):
        """ Create the toolkit-specific control that represents the editor. """

        from enaml.widgets.constraints_widget import ProxyConstraintsWidget

        self.component = self.create_component()

        # We start with an invisible component to avoid flicker. We restore the
        # initial state after the Qt control is parented.
        visible = self.component.visible
        self.component.visible = False

        # Initialize the proxy.
        self.component.initialize()

        # Activate the proxy.
        if not self.component.proxy_is_active:
            self.component.activate_proxy()

        # Fish the Qt control out of the proxy. That's our TaskPane content.
        self.control = self.component.proxy.widget

        # Set the parent
        if parent is not None:
            self.control.setParent(parent)

        # Restore the visibility state
        self.component.visible = visible
        if isinstance(self.component, ProxyConstraintsWidget):
            self.component.proxy.request_relayout()

    def destroy(self):
        """ Destroy the toolkit-specific control that represents the editor.
        """

        control = self.control
        if control is not None:
            control.hide()
            self.component.destroy()
            control.deleteLater()

        self.control = None
        self.component = None



import os
import types
import cStringIO as StringIO
import uuid

import numpy as np
import fs
from fs.opener import opener
import jsonpickle

# Enthought library imports.
from traits.api import HasTraits, Trait, TraitHandler, Int, Any, List, Set, Bool, Event, Dict, Set, Unicode, Property, Str

from omnivore.utils.command import UndoStack
from omnivore.utils.file_guess import FileGuess, FileMetadata
import omnivore.utils.jsonutil as jsonutil

import logging
log = logging.getLogger(__name__)


class DocumentError(RuntimeError):
    pass


class TraitNumpyConverter(TraitHandler):
    """Trait validator to convert bytes to numpy array"""

    def validate(self, object, name, value):
        if type(value) is np.ndarray:
            return value
        elif type(value) is types.StringType:
            return np.fromstring(value, dtype=np.uint8)
        self.error(object, name, value)

    def info(self):
        return '**a string or numpy array**'


class BaseDocument(HasTraits):
    undo_stack = Any

    metadata = Any

    name = Property(Unicode, depends_on='metadata')

    uri = Property(Unicode, depends_on='metadata')

    read_only = Property(Bool, depends_on='metadata')

    document_id = Int(-1)

    uuid = Str

    last_task_id = Str

    baseline_document = Any(transient=True)

    bytes = Trait("", TraitNumpyConverter())

    segments = List

    extra_metadata = Dict

    # Trait events to provide view updating

    undo_stack_changed = Event

    byte_values_changed = Event  # but not the size of the bytes array. That's not handled yet

    byte_style_changed = Event  # only styling info may have changed, not any of the data byte values

    change_count = Int()

    can_revert = Property(Bool, depends_on='metadata')

    permute = Any

    load_error = Str(transient=True)

    #### trait default values

    def _metadata_default(self):
        return FileMetadata(uri="")

    def _undo_stack_default(self):
        return UndoStack()

    def _bytes_default(self):
        return ""

    def _uuid_default(self):
        return str(uuid.uuid4())

    #### trait property getters

    def _get_name(self):
        return self.metadata.name or 'Untitled'

    def _get_uri(self):
        return self.metadata.uri

    def _set_uri(self, uri):
        self.metadata.uri = uri

    def _get_read_only(self):
        return self.metadata.read_only

    def _set_read_only(self, read_only):
        self.metadata.read_only = read_only

    def _get_can_revert(self):
        return self.metadata.uri != ""

    @property
    def menu_name(self):
        if self.uri:
            return "%s (%s)" % (self.name, self.uri)
        return self.name

    @classmethod
    def get_blank(cls):
        return cls(bytes="")

    def __str__(self):
        return "Document(id=%s): %s" % (self.document_id, self.metadata.uri)

    def __len__(self):
        return np.alen(self.bytes)

    def __getitem__(self, val):
        return self.bytes[val]

    @property
    def dirty(self):
        return self.undo_stack.is_dirty()

    def to_bytes(self):
        return self.bytes.tostring()

    def load_permute(self, editor):
        if self.permute:
            self.permute.load(self, editor)

    def filesystem_path(self):
        try:
            fs_, relpath = fs.opener.opener.parse(self.uri)
            if fs_.hassyspath(relpath):
                return fs_.getsyspath(relpath)
        except fs.errors.FSError:
            pass
        return None

    @property
    def bytestream(self):
        return StringIO.StringIO(self.bytes)

    # serialization

    def load_metadata(self, guess):
        extra = self.load_extra_metadata(guess)
        self.restore_extra_from_dict(extra)
        self.extra_metadata = extra

    def load_extra_metadata(self, guess):
        return self.load_filesystem_extra_metadata()

    def load_filesystem_extra_metadata(self):
        """ Find any extra metadata associated with the document, typically
        used to load an extra file off the disk.
        
        If successful, return a dict to be processed by init_extra_metadata
        """
        uri = self.get_filesystem_extra_metadata_uri()
        if uri is None:
            return {}
        try:
            guess = FileGuess(uri)
        except fs.errors.FSError, e:
            log.error("File load error: %s" % str(e))
            return {}
        log.info("Loading metadata file: %s" % uri)
        try:
            b = guess.bytes
            if b.startswith("#"):
                header, b = b.split("\n", 1)
            unserialized = jsonpickle.loads(b)
        except ValueError, e:
            log.error("JSON parsing error for extra metadata in %s: %s" % (uri, str(e)))
            unserialized = {}
        except AttributeError, e:
            log.error("JSON library error: %s: %s" % (uri, str(e)))
            unserialized = {}
            raise DocumentError("JSON library error: % s" % str(e))
        return unserialized

    def get_filesystem_extra_metadata_uri(self):
        """ Get filename of file used to store extra metadata
        """
        return None

    def get_metadata_for(self, task):
        """Return extra metadata for the particular task

        """
        # Each task has its own section in the metadata so they can save stuff
        # without fear of stomping on another task's data. Also, when saving,
        # they can overwrite their task stuff without changing an other task's
        # info so that other task's stuff can be re-saved even if that task
        # wasn't used in this editing session.
        try:
            return self.extra_metadata["task_specific"][task.editor_id]
        except KeyError:
            log.info("%s not in task specific metadata; falling back to old metadata storage" % task.editor_id)

        # For compatibility with pre-1.0 versions of Omnivore which stored
        # metadata for all tasks in the root directory
        return self.extra_metadata

    def init_extra_metadata_dict(self):
        """ Creates new metadata dictionary for metadata to be serialized

        The returned dict includes all the current document properties and all
        the task specific metadata in the originally loaded document.

        The task specific metadata will be replaced by values in the current
        task.
        """
        mdict = {"task_specific": {}}
        if "task_specific" in self.extra_metadata:
            mdict["task_specific"].update(self.extra_metadata["task_specific"])
        self.serialize_extra_to_dict(mdict)
        return mdict

    def store_task_specific_metadata(self, editor, mdict, task_dict):
        # FIXME: should handle all tasks that have changed in this edit
        # session, not just the one that is being saved.
        task_name = editor.task.editor_id
        mdict["task_specific"][task_name] = task_dict
        mdict["last_task_id"] = task_name

    def serialize_extra_to_dict(self, mdict):
        """Save extra metadata to a dict so that it can be serialized
        """
        mdict["document uuid"] = self.uuid
        if self.baseline_document is not None:
            mdict["baseline document"] = self.baseline_document.metadata.uri

    def restore_extra_from_dict(self, e):
        log.debug("restoring extra metadata: %s" % str(e))
        if 'document uuid' in e:
            self.uuid = e['document uuid']
        if 'baseline document' in e:
            try:
                self.load_baseline(e['baseline document'])
            except DocumentError:
                pass
        if 'last_task_id' in e:
            self.last_task_id = e['last_task_id']

    def load_baseline(self, uri, confirm_callback=None):
        if confirm_callback is None:
            confirm_callback = lambda a: True
        try:
            guess = FileGuess(uri)
        except Exception, e:
            log.error("Problem loading baseline file %s: %s" % (uri, str(e)))
            raise DocumentError(str(e))
        bytes = guess.numpy
        difference = len(bytes) - len(self)
        if difference > 0:
            if confirm_callback("Truncate baseline data by %d bytes?" % difference, "Baseline Size Difference"):
                bytes = bytes[0:len(self)]
            else:
                bytes = []
        elif difference < 0:
            if confirm_callback("Pad baseline data with %d zeros?" % (-difference), "Baseline Size Difference"):
                bytes = np.pad(bytes, (0, -difference), "constant", constant_values=0)
            else:
                bytes = []
        if len(bytes) > 0:
            self.init_baseline(guess.metadata, bytes)
        else:
            self.del_baseline()

    def save_to_uri(self, uri, editor, saver=None, save_metadata=True):
        # Have to use a two-step process to write to the file: open the
        # filesystem, then open the file.  Have to open the filesystem
        # as writeable in case this is a virtual filesystem (like ZipFS),
        # otherwise the write to the actual file will fail with a read-
        # only filesystem error.
        if saver is None:
            bytes = self.bytes.tostring()
        else:
            bytes = saver(self, editor)

        if uri.startswith("file://"):
            # FIXME: workaround to allow opening of file:// URLs with the
            # ! character
            uri = uri.replace("file://", "")
        fs, relpath = opener.parse(uri, writeable=True)
        fh = fs.open(relpath, 'wb')
        log.debug("saving to %s" % uri)
        fh.write(bytes)
        fh.close()

        if save_metadata:
            mdict = self.init_extra_metadata_dict()
            task_metadata = dict()
            editor.to_metadata_dict(task_metadata, self)
            self.store_task_specific_metadata(editor, mdict, task_metadata)
            if mdict:
                relpath += ".omnivore"
                log.debug("saving extra metadata to %s" % relpath)
                jsonpickle.set_encoder_options("json", sort_keys=True, indent=4)
                bytes = jsonpickle.dumps(mdict)
                text = jsonutil.collapse_json(bytes)
                header = editor.get_extra_metadata_header()
                fh = fs.open(relpath, 'wb')
                fh.write(header)
                fh.write(text)
                fh.close()

        fs.close()



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: David C. Morrill <dmorrill@enthought.com>
#
#-----------------------------------------------------------------------------
""" Object with views for naming or renaming a user perspective. """


# Enthought library imports.
from traits.api import Bool, HasTraits, Trait, TraitError, Constant
from traitsui.api import View, Item, VGroup


#### Trait definitions ########################################################

def not_empty_string(object, name, value):
    """a not-empty string"""

    if isinstance(value, basestring) and (value.strip() != ''):
        return value

    raise TraitError

# Define a trait which can not be the empty string:
NotEmptyString = Trait('', not_empty_string)


class UserPerspectiveName(HasTraits):
    """ Object with views for naming or renaming a user perspective. """

    ###########################################################################
    # 'UserPerspectiveName' interface.
    ###########################################################################

    # The name of the new user perspective.
    name = NotEmptyString

    # Should the editor area be shown in this perpsective?
    show_editor_area = Bool(True)

    # Help notes when creating a new view.
    new_help = Constant("""Note:
 - The new perspective will initially be empty.
 - Add new views to the perspective by selecting
   them from the 'View' menu.
 - Drag the notebook tabs and splitter bars to
   arrange the views within the perspective.""")

    #### Traits views #########################################################

    new_view = View(
        VGroup(
            VGroup( 'name', 'show_editor_area' ),
            VGroup( '_',
                Item( 'new_help',
                      style = 'readonly' ),
                show_labels = False
            )
        ),
        title   = 'New User Perspective',
        id      = 'envisage.workbench.action.'
                  'new_user_perspective_action.UserPerspectiveName',
        buttons = [ 'OK', 'Cancel' ],
        kind    = 'livemodal',
        width   = 300
    )

    save_as_view = View( 'name',
        title   = 'Save User Perspective As',
        id      = 'envisage.workbench.action.'
                  'save_as_user_perspective_action.UserPerspectiveName',
        buttons = [ 'OK', 'Cancel' ],
        kind    = 'livemodal',
        width   = 300
    )

    rename_view = View( 'name',
        title   = 'Rename User Perspective',
        id      = 'envisage.workbench.action.'
                  'rename_user_perspective_action.UserPerspectiveName',
        buttons = [ 'OK', 'Cancel' ],
        kind    = 'livemodal',
        width   = 300
    )

#### EOF #####################################################################




"""
Test case for ZipFS binary file reading/writing
Passes ok on Linux, fails on Windows (tested: Win7, 64-bit):

AssertionError: ' \r\n' != ' \n'
"""

import unittest
from fs.zipfs import ZipFS
import os

from six import b

class ZipFsBinaryWriteRead(unittest.TestCase):
    test_content = b(chr(32) + chr(10))
    
    def setUp(self):
        self.z = ZipFS('test.zip', 'w')
    
    def tearDown(self):
        try:
            os.remove('test.zip')
        except:
            pass

    def test_binary_write_read(self):
        # GIVEN zipfs
        z = self.z

        # WHEN binary data is written to a test file in zipfs
        f = z.open('test.data', 'wb')
        f.write(self.test_content)
        f.close()
        z.close()

        # THEN the same binary data is retrieved when opened again
        z = ZipFS('test.zip', 'r')
        f = z.open('test.data', 'rb')
        content = f.read()
        f.close()
        z.close()
        self.assertEqual(content, self.test_content)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys

from fs.opener import opener
from fs.commands.runner import Command

from fs.utils import print_fs

class FSTree(Command):
    
    usage = """fstree [OPTION]... [PATH]
Recursively display the contents of PATH in an ascii tree"""
    
    def get_optparse(self):
        optparse = super(FSTree, self).get_optparse()        
        optparse.add_option('-l', '--level', dest='depth', type="int", default=5,
                            help="Descend only LEVEL directories deep (-1 for infinite)", metavar="LEVEL")
        optparse.add_option('-g', '--gui', dest='gui', action='store_true', default=False,
                            help="browse the tree with a gui")
        optparse.add_option('-a', '--all', dest='all', action='store_true', default=False,
                            help="do not hide dot files")        
        optparse.add_option('--dirsfirst', dest='dirsfirst', action='store_true', default=False,
                            help="List directories before files")
        optparse.add_option('-P', dest="pattern", default=None,
                            help="Only list files that match the given pattern")
        optparse.add_option('-d', dest="dirsonly", default=False, action='store_true',
                            help="List directories only")
        return optparse
        
    def do_run(self, options, args):        
        
        if not args:
            args = ['.']
    
        for fs, path, is_dir in self.get_resources(args, single=True):                            
            if not is_dir:
                self.error(u"'%s' is not a dir\n" % path)
                return 1
            fs.cache_hint(True)
            if options.gui:
                from fs.browsewin import browse
                if path:
                    fs = fs.opendir(path)
                browse(fs, hide_dotfiles=not options.all)
            else:
                if options.depth < 0:
                    max_levels = None
                else:
                    max_levels = options.depth
                self.output(self.wrap_dirname(args[0] + '\n'))
                dircount, filecount = print_fs(fs, path or '',
                                               file_out=self.output_file,
                                               max_levels=max_levels,
                                               terminal_colors=self.terminal_colors,
                                               hide_dotfiles=not options.all,
                                               dirs_first=options.dirsfirst,
                                               files_wildcard=options.pattern,
                                               dirs_only=options.dirsonly)
                self.output('\n')
                def pluralize(one, many, count):
                    if count == 1:
                        return '%i %s' % (count, one)
                    else:
                        return '%i %s' % (count, many)
                
                self.output("%s, %s\n" % (pluralize('directory', 'directories', dircount),
                                  pluralize('file', 'files', filecount)))
   
def run():
    return FSTree().run()          
    
if __name__ == "__main__":
    sys.exit(run())
   



import os

import numpy as np

from fs.opener import opener, fsopen
from fs.errors import FSError

from traits.api import HasTraits, Bool, Str, Unicode, Trait, TraitHandler, Property

from omnivore.templates import get_template

import logging
log = logging.getLogger(__name__)


def normalize_uri(uri):
    if uri.startswith("file://"):
        # FIXME: workaround to allow opening of file:// URLs with the
        # ! character
        uri = uri.replace("file://", "")
    if uri:
        fs, relpath = opener.parse(uri)
        if fs.haspathurl(relpath):
            uri = fs.getpathurl(relpath)
        elif fs.hassyspath(relpath):
            abspath = fs.getsyspath(relpath)
            if abspath.startswith("\\\\?\\UNC\\"):
                # Leave long UNC paths as raw paths because there is no
                # accepted way to convert to URI
                uri = abspath
            else:
                if abspath.startswith("\\\\?\\") and len(abspath) < 260:
                    # on windows, pyfilesystem returns extended path notation to
                    # allow paths greater than 256 characters.  If the path is
                    # short, change slashes to normal and remove the prefix
                    abspath = abspath[4:].replace("\\", "/")
                uri = "file://" + abspath
    return uri


class TraitUriNormalizer(TraitHandler):
    """Trait validator to convert bytes to numpy array"""

    def validate(self, object, name, value):
        try:
            uri = normalize_uri(value)
            return uri
        except FSError:
            # Allow the error to be caught when the file is actually opened
            return value
        except:
            self.error(object, name, value)

    def info(self):
        return '**a string or unicode URI**'


class FileMetadata(HasTraits):
    uri = Trait("", TraitUriNormalizer())

    mime = Str(default="application/octet-stream")

    name = Property(Unicode, depends_on='uri')

    read_only = Bool(False)

    def __str__(self):
        return "uri=%s, mime=%s" % (self.uri, self.mime)

    def _get_name(self):
        return os.path.basename(self.uri)


class FileGuess(object):
    """Loads the first part of a file and provides a container for metadata

    """
    # Arbitrary size header, but should be large enough that binary files can
    # be scanned for a signature. Make sure it's not a multiple of 1k for naive
    # parsers that rely on size detection
    head_size = 1024*1024 - 1

    def __init__(self, uri):
        log.debug("Attempting to load %s" % uri)
        self.reload(uri)

    def get_fs(self, uri=None):
        if uri is None:
            uri = self.metadata.uri
        if uri.startswith("file://"):
            # FIXME: workaround to allow opening of file:// URLs with the
            # ! character
            uri = uri.replace("file://", "")
        fs, relpath = opener.parse(uri)
        log.debug("Filesystem: %s" % fs)
        fh = fs.open(relpath, "rb")
        return fh, fs, relpath

    def reload(self, uri=None):
        fh, fs, relpath = self.get_fs(uri)

        # In order to handle arbitrarily sized files, only read the first
        # header bytes.  If the file is bigger, it will be read by the task
        # as needed.
        self._numpy = None
        self.bytes = fh.read(self.head_size)
        fh.close()

        # Use the default mime type until it is recognized
        self.metadata = FileMetadata(uri=uri)
        try:
            self.metadata.read_only = fs.read_only_fs
        except AttributeError:
            pass

        # Release filesystem resources
        fs.close()

    def __str__(self):
        return "guess: metadata: %s, %d bytes available for signature" % (self.metadata, len(self.bytes))

    def get_utf8(self):
        return self.bytes

    @property
    def numpy(self):
        if self._numpy is None:
            self._numpy = np.fromstring(self.bytes, dtype=np.uint8)
        return self._numpy

    def get_metadata(self):
        return self.metadata.clone_traits()

    def get_stream(self):
        fh, fs, relpath = self.get_fs()
        return fh



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------
""" Convenience functions for creating logging handlers etc. """


# Standard library imports.
import logging
from logging.handlers import RotatingFileHandler

# Enthought library imports.
from traits.util.api import deprecated

# Local imports.
from log_queue_handler import LogQueueHandler


# The default logging level.
LEVEL = logging.DEBUG

# The default formatter.
FORMATTER = logging.Formatter('%(levelname)s|%(asctime)s|%(message)s')


class LogFileHandler(RotatingFileHandler):
    """ The default log file handler.
    """

    def __init__(self, path, maxBytes=1000000, backupCount=3, level=None,
        formatter=None):
        RotatingFileHandler.__init__(
            self, path, maxBytes=maxBytes, backupCount=3
        )

        if level is None:
            level = LEVEL
        if formatter is None:
            formatter = FORMATTER
        # Set our default formatter and log level.
        self.setFormatter(formatter)
        self.setLevel(level)

@deprecated('use "LogFileHandler"')
def create_log_file_handler(path, maxBytes=1000000, backupCount=3, level=None,
    formatter=None):
    """ Creates a log file handler.

    This is just a convenience function to make it easy to create the same
    kind of handlers across applications.

    It sets the handler's formatter to the default formatter, and its logging
    level to the default logging level.

    """
    if level is None:
        level = LEVEL
    if formatter is None:
        formatter = FORMATTER

    handler = RotatingFileHandler(
        path, maxBytes=maxBytes, backupCount=backupCount
    )

    handler.setFormatter(formatter)
    handler.setLevel(level)

    return handler


def add_log_queue_handler(logger, level=None, formatter=None):
    """ Adds a queueing log handler to a logger.
    """
    if level is None:
        level = LEVEL
    if formatter is None:
        formatter = FORMATTER

    # Add the handler to the root logger.
    log_queue_handler = LogQueueHandler()
    log_queue_handler.setLevel(level)
    log_queue_handler.setFormatter(formatter)
    logger.addHandler(log_queue_handler)
    return log_queue_handler


#### EOF ######################################################################



""" The nodes used in the extension registry browser tree. """


# Enthought library imports.
from envisage.api import IExtensionPoint, IExtensionRegistry
from traits.api import Undefined
from traitsui.api import TreeNode

# fixme: non-api imports.
from traitsui.value_tree import SingleValueTreeNodeObject
from traitsui.value_tree import value_tree_nodes


class IExtensionRegistryTreeNode(TreeNode):
    """ A tree node for an extension registry. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return True

    def get_children(self, obj):
        """ Get the object's children. """

        extension_points = obj.get_extension_points()

        # Tag the extension registry onto the extension points.
        for extension_point in extension_points:
            extension_point.__extension_registry__ = obj

        return extension_points

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return IExtensionRegistry(obj, Undefined) is obj


class IExtensionPointTreeNode(TreeNode):
    """ A tree node for an extension point. """

    ###########################################################################
    # 'TreeNode' interface.
    ###########################################################################

    def allows_children(self, obj):
        """ Return True if this object allows children. """

        return True

    def get_children(self, obj):
        """ Get the object's children. """

        # fixme: This could be uglier, but I can't work out how ;^)
        index    = 0
        children = []
        for extension in obj.__extension_registry__.get_extensions(obj.id):
            parent = SingleValueTreeNodeObject(value=obj, _index=index)
            children.append(parent.node_for('', extension))
            index += 1

        return children

    def get_label(self, obj):
        """ Get the object's label. """

        return obj.id

    def is_node_for(self, obj):
        """ Return whether this is the node that handles a specified object.

        """

        return IExtensionPoint(obj, Undefined) is obj

    # We override the following methods because 'ExtensionPoint' instances
    # are trait *types* and hence do not actually have traits themselves (i.e.
    # they do not inherit from 'HasTraits'). The default implementations of
    # these methods in 'TreeNode' attempt to call 'on_trait_change' to hook
    # up the listenrs, but obviously, if they don't have traits they don't have
    # 'on_trait_change' either ;^)
    #
    # fixme: If we make this node readonly will these go away?!?
    def when_label_changed(self, obj, callback, remove):
        """ Set up or remove listeners for label changes. """

        return

    def when_children_replaced(self, obj, callback, remove):
        """ Set up or remove listeners for children being replaced. """

        return

    def when_children_changed(self, obj, callback, remove):
        """ Set up or remove listenrs for children being changed. """

        return


extension_registry_browser_tree_nodes = [
    IExtensionRegistryTreeNode(
        auto_open = True,
        label     = '=Extension Points',
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),

    IExtensionPointTreeNode(
        rename    = False,
        copy      = False,
        delete    = False,
        insert    = False,
        menu      = None,
    ),
] + value_tree_nodes

#### EOF ######################################################################



""" Text editor sample task

"""
# Enthought library imports.
from pyface.api import ImageResource, ConfirmationDialog, FileDialog, \
    ImageResource, YES, OK, CANCEL
from pyface.action.api import Action
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import DockPaneToggleGroup, SMenuBar, \
    SMenu, SToolBar, TaskAction, TaskToggleGroup
from traits.api import on_trait_change, Property, Instance

from omnivore.framework.task import FrameworkTask
from styled_text_editor import StyledTextEditor
from preferences import TextEditPreferences


class TextEditTask(FrameworkTask):
    """ A simple task for opening a blank editor.
    """

    new_file_text = ""

    #### Task interface #######################################################

    id = 'omnivore.framework.text_edit_task'
    name = 'Text Editor'

    preferences_helper = TextEditPreferences

    ###########################################################################
    # 'FrameworkTask' interface.
    ###########################################################################

    def get_editor(self, guess=None):
        """ Opens a new empty window
        """
        editor = StyledTextEditor()
        return editor

    ###
    @classmethod
    def can_edit(cls, document):
        return document.metadata.mime.startswith("text/")

    @classmethod
    def get_match_score(cls, document):
        """Return a number based on how good of a match this task is to the
        incoming Document.
        
        0 = generic match
        ...
        10 = absolute match
        """
        if document.metadata.mime.startswith("text/"):
            # Leave 10 for exact matches, like text/html for HtmlView mode
            return 9
        return 0



#!/usr/bin/env python
from fs.commands.runner import Command
import sys

class FSMkdir(Command):
    
    usage = """fsmkdir [PATH]
Make a directory"""

    version = "1.0"
    
    def do_run(self, options, args):
                
        for fs_url in args:                    
            self.open_fs(fs_url, create_dir=True)                        
    
def run():
    return FSMkdir().run()
    
if __name__ == "__main__":
    sys.exit(run())
        


import wx

from traits.api import Float, Any, Str, Trait
from traitsui.editors.api import RangeEditor
from traitsui.wx.editor import Editor
from traitsui.wx.helper import TraitsUIPanel, Slider


class _BoundsEditor(Editor):

    evaluate = Any

    min = Any
    max = Any
    low = Any
    high = Any
    format = Str

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if not factory.low_name:
            self.low = factory.low
            self.min = self.low

        if not factory.high_name:
            self.high = factory.high
            self.max = self.high

        self.max = factory.max
        self.min = factory.min

        self.format = factory.format

        self.evaluate = factory.evaluate
        self.sync_value(factory.evaluate_name, 'evaluate', 'from')

        self.sync_value(factory.low_name, 'low', 'both')
        self.sync_value(factory.high_name, 'high', 'both')

        self.control = panel = TraitsUIPanel(parent, -1)
        sizer = wx.FlexGridSizer(2, 3, 0, 0)

        # low text box
        self._label_lo = wx.TextCtrl(panel, -1, self.format % self.low,
                                     size=wx.Size(56, 20),
                                     style=wx.TE_PROCESS_ENTER)
        sizer.Add(self._label_lo, 0, wx.ALIGN_CENTER)
        wx.EVT_TEXT_ENTER(
            panel,
            self._label_lo.GetId(),
            self.update_low_on_enter)
        wx.EVT_KILL_FOCUS(self._label_lo, self.update_low_on_enter)

        # low slider
        self.control.lslider = Slider(panel, -1, 0, 0, 10000,
                                      size=wx.Size(100, 20),
                                      style=wx.SL_HORIZONTAL | wx.SL_AUTOTICKS)
        self.control.lslider.SetValue(self._convert_to_slider(self.low))
        self.control.lslider.SetTickFreq(1000, 1)
        self.control.lslider.SetPageSize(1000)
        self.control.lslider.SetLineSize(100)
        wx.EVT_SCROLL(self.control.lslider, self.update_object_on_scroll)
        sizer.Add(self.control.lslider, 1, wx.EXPAND)
        sizer.AddStretchSpacer(0)

        # high slider
        sizer.AddStretchSpacer(0)
        self.control.rslider = Slider(
            panel, -1, self._convert_to_slider(
                self.high), 0, 10000, size=wx.Size(
                100, 20), style=wx.SL_HORIZONTAL | wx.SL_AUTOTICKS)
        self.control.rslider.SetTickFreq(1000, 1)
        self.control.rslider.SetPageSize(1000)
        self.control.rslider.SetLineSize(100)
        wx.EVT_SCROLL(self.control.rslider, self.update_object_on_scroll)
        sizer.Add(self.control.rslider, 1, wx.EXPAND)

        # high text box
        self._label_hi = wx.TextCtrl(panel, -1, self.format % self.high,
                                     size=wx.Size(56, 20),
                                     style=wx.TE_PROCESS_ENTER)
        sizer.Add(self._label_hi, 0, wx.ALIGN_CENTER)
        wx.EVT_TEXT_ENTER(
            panel,
            self._label_hi.GetId(),
            self.update_high_on_enter)
        wx.EVT_KILL_FOCUS(self._label_hi, self.update_high_on_enter)

        self.set_tooltip(self.control.lslider)
        self.set_tooltip(self.control.rslider)
        self.set_tooltip(self._label_lo)
        self.set_tooltip(self._label_hi)

        # Set-up the layout:
        panel.SetSizerAndFit(sizer)

    def update_low_on_enter(self, event):
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        try:
            try:
                low = eval(unicode(self._label_lo.GetValue()).strip())
                if self.evaluate is not None:
                    low = self.evaluate(low)
            except Exception as ex:
                low = self.low
                self._label_lo.SetValue(self.format % self.low)

            if not self.factory.is_float:
                low = int(low)

            if low > self.high:
                low = self.high - self._step_size()
                self._label_lo.SetValue(self.format % low)

            self.control.lslider.SetValue(self._convert_to_slider(low))
            self.low = low
        except:
            pass

    def update_high_on_enter(self, event):
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        try:
            try:
                high = eval(unicode(self._label_hi.GetValue()).strip())
                if self.evaluate is not None:
                    high = self.evaluate(high)
            except:
                high = self.high
                self._label_hi.SetValue(self.format % self.high)

            if not self.factory.is_float:
                high = int(high)

            if high < self.low:
                high = self.low + self._step_size()
                self._label_hi.SetValue(self.format % high)

            self.control.rslider.SetValue(self._convert_to_slider(high))
            self.high = high
        except:
            pass

    def update_object_on_scroll(self, evt):
        low = self._convert_from_slider(self.control.lslider.GetValue())
        high = self._convert_from_slider(self.control.rslider.GetValue())

        if low >= high:
            if evt.Position == self.control.lslider.GetValue():
                low = self.high - self._step_size()
            else:
                high = self.low + self._step_size()

        if self.factory.is_float:
            self.low = low
            self.high = high
        else:
            self.low = int(low)
            self.high = int(high)

            # update the sliders to the int values or the sliders
            # will jiggle
            self.control.lslider.SetValue(self._convert_to_slider(low))
            self.control.rslider.SetValue(self._convert_to_slider(high))

    def update_editor(self):
        return

    def _check_max_and_min(self):
        # check if max & min have been defined:
        if self.max is None:
            self.max = self.high
        if self.min is None:
            self.min = self.low

    def _step_size(self):
        slider_delta = self.control.lslider.GetMax() - self.control.lslider.GetMin()
        range_delta = self.max - self.min

        return float(range_delta) / slider_delta

    def _convert_from_slider(self, slider_val):
        self._check_max_and_min()
        return self.min + slider_val * self._step_size()

    def _convert_to_slider(self, value):
        self._check_max_and_min()
        return self.control.lslider.GetMin() + (value - self.min) / self._step_size()

    def _low_changed(self, low):
        if self.control is None:
            return
        if self._label_lo is not None:
            self._label_lo.SetValue(self.format % low)

        self.control.lslider.SetValue(self._convert_to_slider(low))

    def _high_changed(self, high):
        if self.control is None:
            return
        if self._label_hi is not None:
            self._label_hi.SetValue(self.format % high)

        self.control.rslider.SetValue(self._convert_to_slider(self.high))


class BoundsEditor(RangeEditor):

    min = Trait(None, Float)
    max = Trait(None, Float)

    def _get_simple_editor_class(self):
        return _BoundsEditor

    def _get_custom_editor_class(self):
        return _BoundsEditor



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   08/23/2008
#
#------------------------------------------------------------------------------

""" A base class for creating custom Traits UI windows.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import HasPrivateTraits, Instance, Property

from helper \
    import init_wx_handlers, BufferDC

#-------------------------------------------------------------------------
#  'UIWindow' class:
#-------------------------------------------------------------------------


class UIWindow(HasPrivateTraits):
    """ A base class for creating custom Traits UI windows.
    """

    # The wx.Window associated with this custom window:
    control = Instance(wx.Window)

    # The initial size of the window:
    size = Instance(wx.Size, (-1, -1))

    # The current width of the window:
    width = Property

    # The current height of the window:
    height = Property

    #-- Public Methods -------------------------------------------------------

    def __init__(self, parent, **traits):
        """ Creates and initializes the window.
        """
        super(UIWindow, self).__init__(**traits)
        self.control = wx.Window(parent, -1,
                                 size=self.size,
                                 style=wx.FULL_REPAINT_ON_RESIZE)
        init_wx_handlers(self.control, self)

    def refresh(self, x=None, y=None, dx=None, dy=None):
        """ Refreshes the contents of the window.
        """
        if self.control is not None:
            if x is None:
                self.control.Refresh()
            else:
                self.control.Refresh(x, y, dx, dy)

    def capture(self):
        """ Capture the mouse.
        """
        self.control.CaptureMouse()

    def release(self):
        """ Release the mouse.
        """
        self.control.ReleaseMouse()

    #-- wxPython Event Handlers ----------------------------------------------

    def _erase_background(self, event):
        """ Never, ever, do anything in this handler.
        """
        pass

    def _paint(self, event):
        """ Paints the contents of the window.
        """
        dc = BufferDC(self.control)
        self._paint_dc(dc)
        dc.copy()

    def _paint_dc(self, dc):
        """ This method should be overridden by sub-classes to do the actual
            window painting.
        """
        pass

    #-- Property Implementations ---------------------------------------------

    def _get_width(self):
        return self.control.GetClientSize()[0]

    def _set_width(self, width):
        self.control.SetSize(width, self.height)

    def _get_height(self):
        return self.control.GetClientSize()[1]

    def _set_height(self, height):
        self.control.SetSize(self.width, height)



from dict_context_adapter import DictContextAdapter
from dict_context_adapter_factory import DictContextAdapterFactory
from instance_context_adapter import InstanceContextAdapter
from instance_context_adapter_factory import InstanceContextAdapterFactory
from list_context_adapter import ListContextAdapter
from list_context_adapter_factory import ListContextAdapterFactory
from trait_list_context_adapter import TraitListContextAdapter
from trait_list_context_adapter_factory import TraitListContextAdapterFactory
from tuple_context_adapter import TupleContextAdapter
from tuple_context_adapter_factory import TupleContextAdapterFactory

from trait_dict_context_adapter import TraitDictContextAdapter



#------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..api import HasTraits, Str, Undefined, ReadOnly, Float


class Foo(HasTraits):
    name = Str()
    original_name = ReadOnly

    bar = Str
    baz = Float

    def _name_changed(self):
        if self.original_name is Undefined:
            self.original_name = self.name


class Bar(HasTraits):
    name = Str(Undefined)


class UndefinedTestCase(unittest.TestCase):
    def test_initial_value(self):
        b = Bar()
        self.assertEqual(b.name, Undefined)
        return

    def test_name_change(self):
        b = Bar()
        b.name = 'first'
        self.assertEqual(b.name, 'first')
        return

    def test_read_only_write_once(self):
        f = Foo()

        self.assertEqual(f.name, '')
        self.assertIs(f.original_name, Undefined)

        f.name = 'first'
        self.assertEqual(f.name, 'first')
        self.assertEqual(f.original_name, 'first')

        f.name = 'second'
        self.assertEqual(f.name, 'second')
        self.assertEqual(f.original_name, 'first')

        return

    def test_read_only_write_once_from_constructor(self):
        f = Foo(name='first')

        f.name = 'first'
        self.assertEqual(f.name, 'first')
        self.assertEqual(f.original_name, 'first')

        f.name = 'second'
        self.assertEqual(f.name, 'second')
        self.assertEqual(f.original_name, 'first')

        return

### EOF #######################################################################



#-------------------------------------------------------------------------------
#
#  Trait definitions useful when creating templatizable classes.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Trait definitions useful when creating templatizable classes.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Instance, Int, Float, Str, List, Bool, Range, TraitType, Undefined

from itemplate_data_source \
    import ITemplateDataSource

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# Data source template trait:
TDataSource = Instance( ITemplateDataSource, template = 'datasource' )

# Templatizable object trait:
def TInstance ( *args, **kw ):
    kw[ 'template' ] = 'instance'
    return Instance( *args, **kw )

# A list of templatizable object traits:
def TList( *args, **kw ):
    kw[ 'template' ] = 'instance'
    return List( *args, **kw )

# Simple, copyable template traits:
TInt   = Int(   template = 'copy' )
TFloat = Float( template = 'copy' )
TStr   = Str(   template = 'copy' )
TBool  = Bool(  template = 'copy' )

def TRange ( *args, **kw ):
    kw[ 'template' ] = 'copy'
    return Range( *args, **kw )

def TEnum ( *args, **kw ):
    kw[ 'template' ] = 'copy'
    return Enum( *args, **kw )

#-------------------------------------------------------------------------------
#  'TDerived' trait:
#-------------------------------------------------------------------------------

class TDerived ( TraitType ):
    """ Defines a trait property for handling attributes whose value depends
        upon the templatizable traits of an object.

        Note that the implementation of the *activate_template* method is not
        allowed to read the value of any **TDerived** traits, it may only set
        them.
    """

    # Base trait metadata:
    metadata = { 'template':  'derived',
                 'transient': True }

    def get ( self, object, name ):
        name += '_'
        value = object.__dict__.get( name, Undefined )
        if value is not Undefined:
            return value

        object.activate_template()

        value = object.__dict__.get( name, Undefined )
        if value is not Undefined:
            return value

        object.__dict__[ name ] = None
        return None

    def set ( self, object, name, value ):
        xname = name + '_'
        old   = object.__dict__.get( xname, Undefined )
        if old is not value:
            object.__dict__[ xname ] = value
            object.trait_property_changed( name, old, value )




#!/usr/bin/env python

"""
This builds a kml/kmz file for Google earth from
a series of GNOME moss files

In this version the template and pngs are embedded.

"""

from mako.template import Template
import zipfile, base64
import os,sys
import ReadMoss
import Moss2Kmz
import MossKmlTemplates


def MossTime2GoogleTime(mossTime, gmtOffset):
        # mossTime = spillInfo["VALIDFOR"] # 15:00, 08/17/09
        #print mossTime
        timeStr,dateStr = mossTime.split(",")
        dateStr = dateStr.strip()
        monthStr,dayStr,yearStr = dateStr.split("/")
        year = int(yearStr)
        if gmtOffset == 0 :
            offsetStr = "Z"
        elif gmtOffset  > 0 :
            offsetStr = "+02d:00"%(gmtOffset)
        else :
            offsetStr = "-02d:00"%(gmtOffset)

        if year < 50 :
            fourDigitYearStr = "20"+yearStr
        elif year <= 99 :
            fourDigitYearStr = "19"+yearStr
        else:  # already a 4 digit year
            fourDigitYearStr = yearStr

        googleStr = "%s-%s-%sT%s:00%s" % (fourDigitYearStr,monthStr,dayStr,timeStr,offsetStr)
        #print googleStr
        return googleStr


class MossSeriesItem():
    def __init__(self,beginTime,endTime,setsOfObjects):

        self.beginTime = beginTime
        self.endTime = endTime
        self.setsOfObjects = setsOfObjects




def buildkmzSeries(mossSeriesFilename,printDiagnostic = False):
    Name = os.path.basename(mossSeriesFilename) # mossfilename may include the path, so get just the file name

    assert Name.endswith("000") #require that the last three chars of the name be "000"
    baseName = Name[:-3] # chop off the last three characters
    basePathAndName = mossSeriesFilename[:-3]

    seriesDescription = GetDescription(mossSeriesFilename)

    outfilename = basePathAndName+".kmz"
    print "rendering: %s"%outfilename

    BestGuess_Series = []
    Uncertainty_Series = []
    Polygon_Series = []

    for i in range(0,999) :
        mossfilename =  "%s%03d"%(basePathAndName,i)
        print "Processing",mossfilename

        extension = ".ms3"
        if not os.path.exists(mossfilename + extension):
            break # no more moss files in the series

        bestGuessLEs,uncertaintyLEs = ReadMoss.ReadMossLeFiles(mossfilename)
        description = Moss2Kmz.GetDescription(mossfilename)
        spillInfo = ReadMoss.ReadSpillInfo(mossfilename)

        fromLogo = spillInfo["FROMLOGO"] # usually either a gnome or noaa logo
        #if os.path.exists(os.path.join(os.path.dirname(mossSeriesFilename) + "scalebar.png")):
        #fromLogo = "scalebar.png" # try this
        fromLogo = "" # turn this off for now

        LEs_Floating = ReadMoss.FilterLEsByStatus(bestGuessLEs,"INWATER")
        LEs_Beached = ReadMoss.FilterLEsByStatus(bestGuessLEs,"ONBEACH")

        uncertaintyLEs_Floating = ReadMoss.FilterLEsByStatus(uncertaintyLEs,"INWATER")
        uncertaintyLEs_Beached = ReadMoss.FilterLEsByStatus(uncertaintyLEs,"ONBEACH")

        landPolygons,forecastHeavyPolygons,forecastMediumPolygons,forecastLightPolygons,uncertaintyPolygons = ReadMoss.ReadMossPolygons(mossfilename)

        #print description

        mytemplate = Template(LE_Template)

        # ( Z = zulu, or can have +-hh:mm for conversion to UTC
        mossTime = spillInfo["VALIDFOR"] #  example: 15:00, 08/17/09
        gmtOffset = 0 # code goes here
        beginTime = MossTime2GoogleTime(mossTime, gmtOffset)
        endTime = None
        #print "mossTime %s\n beginTime %s" %(mossTime,beginTime)


        altitude = 0.1 # raise the best guess LEs up above the water so that they are always on top of the uncertainty LEs
        LE_Sets = [
                   (LEs_Floating,"%s - %s" %(mossTime,"Floating Splots (Best Guess)"),"#YellowDotIcon",altitude),
                   (LEs_Beached,"%s - %s" %(mossTime,"Beached Splots (Best Guess)"),"#YellowXIcon",altitude)
                   ]

        altitude = 0 # on the water surface
        UncertaintyLE_Sets = [
                   (uncertaintyLEs_Floating,"%s - %s" %(mossTime,"Uncertainty Floating Splots"),"#RedDotIcon",altitude),
                   (uncertaintyLEs_Beached,"%s - %s" %(mossTime,"Uncertainty Beached Splots"),"#RedXIcon",altitude)
                   ]

        Polygon_Sets = [
                        (forecastHeavyPolygons,"%s - %s" %(mossTime,"Forecast Heavy"),"#HeavyStyle",altitude),
                       (forecastMediumPolygons,"%s - %s" %(mossTime,"Forecast Medium"),"#MediumStyle",altitude),
                       (forecastLightPolygons,"%s - %s" %(mossTime,"Forecast Light"),"#LightStyle",altitude),
                       (uncertaintyPolygons,"%s - %s" %(mossTime,"Uncertainty"),"#UncertaintyStyle",altitude)
                      ]

        seriesItem = MossSeriesItem(beginTime,endTime,LE_Sets)
        BestGuess_Series.append(seriesItem)

        seriesItem = MossSeriesItem(beginTime,endTime,UncertaintyLE_Sets)
        Uncertainty_Series.append(seriesItem)

        seriesItem = MossSeriesItem(beginTime,endTime,Polygon_Sets)
        Polygon_Series.append(seriesItem)




    if (
        len(BestGuess_Series) == 0
        and len(Uncertainty_Series) == 0
        and len(Polygon_Series) == 0
        ):
        print "no files processed"
        return

    # set the end times
    for timeSeries in [BestGuess_Series,Uncertainty_Series,Polygon_Series]:
        for i in range(0,len(timeSeries)-1):
            timeSeries[i].endTime = timeSeries[i+1].beginTime


    kmlstring = mytemplate.render(
                                  BestGuess_Series =BestGuess_Series,
                                  Uncertainty_Series =Uncertainty_Series,
                                  Polygon_Series =Polygon_Series,
                                  name=Name,
                                  description = description,
                                  fromLogo = fromLogo)

    kmzfile = zipfile.ZipFile(outfilename, 'w', compression=zipfile.ZIP_DEFLATED)

    kmzfile.writestr('dot.png', base64.b64decode(MossKmlTemplates.DotData))
    kmzfile.writestr('x.png', base64.b64decode(MossKmlTemplates.XData))
    kmzfile.writestr('scalebar.png', base64.b64decode(MossKmlTemplates.ScaleBarData))

    kmzfile.writestr(Name+".kml", kmlstring)
    kmzfile.close()

    if printDiagnostic: print kmlstring

    print "done"


def GetDescription(mossfilename):
    '''
    0, SPILLID:
    0, FROM:
    0, CONTACT:
    0, ISSUED: 13:20, 7/24/09
    0, VALIDFOR: 06:00, 7/17/09
    0, ADDLEDATA:
    0, CAVEAT: This trajectory was produced by GNOME (General NOAA Operational Modeling Environment),
    0, CAVEAT: and should be used for educational and planning purposes only--not for a real response. In the
    0, CAVEAT: event of an oil or chemical spill in U.S. waters, contact the U.S. Coast Guard National Response
    0, CAVEAT: Center at 1-800-424-8802.
    0, CAVEAT:
    0, FROMLOGO: gnome.bmp
    0, LIGHTHEAVYRANGE: 3,5
    '''

    spillInfo = ReadMoss.ReadSpillInfo(mossfilename)
    if spillInfo == None:  return None


    str = ""

    for label,key in [("Valid for","VALIDFOR"),
                      ("From","FROM"),
                      ("Contact","CONTACT"),
                       ("Issued","ISSUED")]:
        if spillInfo[key]: str = str+"<b>%s:</b> %s<br>\n"%(label,spillInfo[key])

    if spillInfo["CAVEAT"]:
        str = str + "<br>\n"+ spillInfo["CAVEAT"]

    return str


#############################################
### the templates:



################################
################################

LE_Template = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
  <Document>
    <name>${name}</name>
    <open>1</open>
      % if description :
      <description><![CDATA[${description}]]></description>
      % endif

""" + (
   MossKmlTemplates.LeStyles +
   MossKmlTemplates.ContourStyles
)  +  """

% for seriesType, series in [("Best Guess",BestGuess_Series),("Uncertainty",Uncertainty_Series)]:
% if series and len(series) > 0 :
 <Folder>
 <name>${seriesType}</name>
% for seriesItem in series:
% for LEs,LEsname,LEsstyle,altitude in seriesItem.setsOfObjects:
   % if LEs :
     <Placemark>
      <name>${LEsname}</name>
      <styleUrl>${LEsstyle}</styleUrl>
      %if seriesItem.beginTime or seriesItem.endTime:
      <TimeSpan id="ID">
          % if seriesItem.beginTime :
          <begin>${seriesItem.beginTime}</begin>     <!-- kml:dateTime -->
          % endif
          % if seriesItem.endTime :
          <end>${seriesItem.endTime}</end>         <!-- kml:dateTime -->
          % endif
      </TimeSpan>
      % endif
      <MultiGeometry>
        % for le in LEs:
             <Point>
                 % if altitude == 0 :
                     <coordinates>${le.longitudeStr},${le.latitudeStr}</coordinates>
                 % else :
                     <altitudeMode>relativeToGround</altitudeMode>
                     <coordinates>${le.longitudeStr},${le.latitudeStr},${altitude}</coordinates>
                 % endif
             </Point>
        % endfor

      </MultiGeometry>
     </Placemark>
    %endif
% endfor
% endfor
 </Folder>
% endif
% endfor

""" + (
   MossKmlTemplates.LogoOverlay
)  +  """

  </Document>
</kml>
"""



if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        filePath = sys.argv[1]
        ext = filePath.split('.')[-1]
        if len(ext) == 3 and ext[:2].lower() == 'ms':
            # this looks like it has a moss extension, so strip it.
            filePath = filePath[:-4]
        buildkmzSeries(filePath)
    else:
        '''
        If called with no argument, run the sample files.
        '''
        print "***************"
        print "No argument provided.  Processing the sample files..."
        print "***************"
        os.chdir(os.path.abspath(os.path.dirname(sys.argv[0])))
        for filePath in ["samplemossfiles/jerryseries/jerryseries000"]:
            buildkmzSeries(filePath)





from __future__ import unicode_literals

from fs import iotools

import io
import unittest
from os.path import dirname, join, abspath

try:
    unicode
except NameError:
    unicode = str


class OpenFilelike(object):
    def __init__(self, make_f):
        self.make_f = make_f

    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        return self.make_f()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.f.close()


class TestIOTools(unittest.TestCase):

    def get_bin_file(self):
        path = join(dirname(abspath(__file__)), 'data/UTF-8-demo.txt')
        return io.open(path, 'rb')

    def test_make_stream(self):
        """Test make_stream"""
        with self.get_bin_file() as f:
            text = f.read()
            self.assert_(isinstance(text, bytes))

        with self.get_bin_file() as f:
            with iotools.make_stream("data/UTF-8-demo.txt", f, 'rt') as f2:
                text = f2.read()
                self.assert_(isinstance(text, unicode))

    def test_decorator(self):
        """Test filelike_to_stream decorator"""
        o = OpenFilelike(self.get_bin_file)
        with o.open('file', 'rb') as f:
            text = f.read()
            self.assert_(isinstance(text, bytes))

        with o.open('file', 'rt') as f:
            text = f.read()
            self.assert_(isinstance(text, unicode))




from pyface.gui import GUI
from traits.api import HasTraits, Instance, Int, List, Str

from traitsui.api import Item, ObjectColumn, TableEditor, View
from traitsui.tests._tools import (
    skip_if_not_qt4, press_ok_button,
    skip_if_null, store_exceptions_on_all_threads)


class ListItem(HasTraits):
    """ Items to visualize in a table editor """
    value = Str
    other_value = Int


class ObjectList(HasTraits):
    values = List(Instance(ListItem))


simple_view = View(
    Item(
        'values',
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name='value'),
                ObjectColumn(name='other_value'),
            ],
        )
    ),
    buttons=['OK'],
)


@skip_if_null
def test_table_editor():
    gui = GUI()
    object_list = ObjectList(
        values=[ListItem(value=str(i**2)) for i in range(10)]
    )

    with store_exceptions_on_all_threads():
        ui = object_list.edit_traits(view=simple_view)
        gui.process_events()
        press_ok_button(ui)
        gui.process_events()


@skip_if_not_qt4
def test_progress_column():
    from traitsui.extras.progress_column import ProgressColumn
    progress_view = View(
        Item(
            'values',
            show_label=False,
            editor=TableEditor(
                columns=[
                    ObjectColumn(name='value'),
                    ProgressColumn(name='other_value'),
                ],
            )
        ),
        buttons=['OK'],
    )
    gui = GUI()
    object_list = ObjectList(
        values=[ListItem(value=str(i**2)) for i in range(10)]
    )

    with store_exceptions_on_all_threads():
        ui = object_list.edit_traits(view=progress_view)
        gui.process_events()
        press_ok_button(ui)
        gui.process_events()



#!/usr/bin/env python
"""

  fs.tests:  testcases for the fs module

"""

from __future__ import with_statement

#  Send any output from the logging module to stdout, so it will
#  be captured by nose and reported appropriately
import sys
import logging
logging.basicConfig(level=logging.ERROR, stream=sys.stdout)

from fs.base import *
from fs.path import *
from fs.errors import *
from fs.filelike import StringIO

import datetime
import unittest
import os
import os.path
import pickle
import random
import copy

import time
try:
    import threading
except ImportError:
    import dummy_threading as threading

import six
from six import PY3, b


class FSTestCases(object):
    """Base suite of testcases for filesystem implementations.

    Any FS subclass should be capable of passing all of these tests.
    To apply the tests to your own FS implementation, simply use FSTestCase
    as a mixin for your own unittest.TestCase subclass and have the setUp
    method set self.fs to an instance of your FS implementation.

    NB. The Filesystem being tested must have a capacity of at least 3MB.

    This class is designed as a mixin so that it's not detected by test
    loading tools such as nose.
    """

    def check(self, p):
        """Check that a file exists within self.fs"""
        return self.fs.exists(p)

    def test_invalid_chars(self):
        """Check paths validate ok"""
        # Will have to be overriden selectively for custom validepath methods
        self.assertEqual(self.fs.validatepath(''), None)
        self.assertEqual(self.fs.validatepath('.foo'), None)
        self.assertEqual(self.fs.validatepath('foo'), None)
        self.assertEqual(self.fs.validatepath('foo/bar'), None)
        self.assert_(self.fs.isvalidpath('foo/bar'))

    def test_tree(self):
        """Test tree print"""
        self.fs.makedir('foo')
        self.fs.createfile('foo/bar.txt')
        self.fs.tree()

    def test_meta(self):
        """Checks getmeta / hasmeta are functioning"""
        # getmeta / hasmeta are hard to test, since there is no way to validate
        # the implementation's response
        meta_names = ["read_only",
                      "network",
                      "unicode_paths"]
        stupid_meta = 'thismetashouldnotexist!"r$$%^&&*()_+'
        self.assertRaises(NoMetaError, self.fs.getmeta, stupid_meta)
        self.assertFalse(self.fs.hasmeta(stupid_meta))
        self.assertEquals(None, self.fs.getmeta(stupid_meta, None))
        self.assertEquals(3.14, self.fs.getmeta(stupid_meta, 3.14))
        for meta_name in meta_names:
            try:
                meta = self.fs.getmeta(meta_name)
                self.assertTrue(self.fs.hasmeta(meta_name))
            except NoMetaError:
                self.assertFalse(self.fs.hasmeta(meta_name))

    def test_root_dir(self):
        self.assertTrue(self.fs.isdir(""))
        self.assertTrue(self.fs.isdir("/"))
        # These may be false (e.g. empty dict) but mustn't raise errors
        self.fs.getinfo("")
        self.assertTrue(self.fs.getinfo("/") is not None)

    def test_getsyspath(self):
        try:
            syspath = self.fs.getsyspath("/")
        except NoSysPathError:
            pass
        else:
            self.assertTrue(isinstance(syspath, unicode))
        syspath = self.fs.getsyspath("/", allow_none=True)
        if syspath is not None:
            self.assertTrue(isinstance(syspath, unicode))

    def test_debug(self):
        str(self.fs)
        repr(self.fs)
        self.assert_(hasattr(self.fs, 'desc'))

    def test_open_on_directory(self):
        self.fs.makedir("testdir")
        try:
            f = self.fs.open("testdir")
        except ResourceInvalidError:
            pass
        except Exception:
            raise
            ecls = sys.exc_info()[0]
            assert False, "%s raised instead of ResourceInvalidError" % (ecls,)
        else:
            f.close()
            assert False, "ResourceInvalidError was not raised"

    def test_writefile(self):
        self.assertRaises(ResourceNotFoundError, self.fs.open, "test1.txt")
        f = self.fs.open("test1.txt", "wb")
        f.write(b("testing"))
        f.close()
        self.assertTrue(self.check("test1.txt"))
        f = self.fs.open("test1.txt", "rb")
        self.assertEquals(f.read(), b("testing"))
        f.close()
        f = self.fs.open("test1.txt", "wb")
        f.write(b("test file overwrite"))
        f.close()
        self.assertTrue(self.check("test1.txt"))
        f = self.fs.open("test1.txt", "rb")
        self.assertEquals(f.read(), b("test file overwrite"))
        f.close()

    def test_createfile(self):
        test = b('now with content')
        self.fs.createfile("test.txt")
        self.assert_(self.fs.exists("test.txt"))
        self.assertEqual(self.fs.getcontents("test.txt", "rb"), b(''))
        self.fs.setcontents("test.txt", test)
        self.fs.createfile("test.txt")
        self.assertEqual(self.fs.getcontents("test.txt", "rb"), test)
        self.fs.createfile("test.txt", wipe=True)
        self.assertEqual(self.fs.getcontents("test.txt", "rb"), b(''))

    def test_readline(self):
        text = b"Hello\nWorld\n"
        self.fs.setcontents('a.txt', text)
        with self.fs.open('a.txt', 'rb') as f:
            line = f.readline()
        self.assertEqual(line, b"Hello\n")

    def test_setcontents(self):
        #  setcontents() should accept both a string...
        self.fs.setcontents("hello", b("world"))
        self.assertEquals(self.fs.getcontents("hello", "rb"), b("world"))
        #  ...and a file-like object
        self.fs.setcontents("hello", StringIO(b("to you, good sir!")))
        self.assertEquals(self.fs.getcontents(
            "hello", "rb"), b("to you, good sir!"))
        #  setcontents() should accept both a string...
        self.fs.setcontents("hello", b("world"), chunk_size=2)
        self.assertEquals(self.fs.getcontents("hello", "rb"), b("world"))
        #  ...and a file-like object
        self.fs.setcontents("hello", StringIO(
            b("to you, good sir!")), chunk_size=2)
        self.assertEquals(self.fs.getcontents(
            "hello", "rb"), b("to you, good sir!"))
        self.fs.setcontents("hello", b(""))
        self.assertEquals(self.fs.getcontents("hello", "rb"), b(""))

    def test_setcontents_async(self):
        #  setcontents() should accept both a string...
        self.fs.setcontents_async("hello", b("world")).wait()
        self.assertEquals(self.fs.getcontents("hello", "rb"), b("world"))
        #  ...and a file-like object
        self.fs.setcontents_async("hello", StringIO(
            b("to you, good sir!"))).wait()
        self.assertEquals(self.fs.getcontents("hello"), b("to you, good sir!"))
        self.fs.setcontents_async("hello", b("world"), chunk_size=2).wait()
        self.assertEquals(self.fs.getcontents("hello", "rb"), b("world"))
        #  ...and a file-like object
        self.fs.setcontents_async("hello", StringIO(
            b("to you, good sir!")), chunk_size=2).wait()
        self.assertEquals(self.fs.getcontents(
            "hello", "rb"), b("to you, good sir!"))

    def test_isdir_isfile(self):
        self.assertFalse(self.fs.exists("dir1"))
        self.assertFalse(self.fs.isdir("dir1"))
        self.assertFalse(self.fs.isfile("a.txt"))
        self.fs.setcontents("a.txt", b(''))
        self.assertFalse(self.fs.isdir("dir1"))
        self.assertTrue(self.fs.exists("a.txt"))
        self.assertTrue(self.fs.isfile("a.txt"))
        self.assertFalse(self.fs.exists("a.txt/thatsnotadir"))
        self.fs.makedir("dir1")
        self.assertTrue(self.fs.isdir("dir1"))
        self.assertTrue(self.fs.exists("dir1"))
        self.assertTrue(self.fs.exists("a.txt"))
        self.fs.remove("a.txt")
        self.assertFalse(self.fs.exists("a.txt"))

    def test_listdir(self):
        def check_unicode(items):
            for item in items:
                self.assertTrue(isinstance(item, unicode))
        self.fs.setcontents(u"a", b(''))
        self.fs.setcontents("b", b(''))
        self.fs.setcontents("foo", b(''))
        self.fs.setcontents("bar", b(''))
        # Test listing of the root directory
        d1 = self.fs.listdir()
        self.assertEqual(len(d1), 4)
        self.assertEqual(sorted(d1), [u"a", u"b", u"bar", u"foo"])
        check_unicode(d1)
        d1 = self.fs.listdir("")
        self.assertEqual(len(d1), 4)
        self.assertEqual(sorted(d1), [u"a", u"b", u"bar", u"foo"])
        check_unicode(d1)
        d1 = self.fs.listdir("/")
        self.assertEqual(len(d1), 4)
        check_unicode(d1)
        # Test listing absolute paths
        d2 = self.fs.listdir(absolute=True)
        self.assertEqual(len(d2), 4)
        self.assertEqual(sorted(d2), [u"/a", u"/b", u"/bar", u"/foo"])
        check_unicode(d2)
        # Create some deeper subdirectories, to make sure their
        # contents are not inadvertantly included
        self.fs.makedir("p/1/2/3", recursive=True)
        self.fs.setcontents("p/1/2/3/a", b(''))
        self.fs.setcontents("p/1/2/3/b", b(''))
        self.fs.setcontents("p/1/2/3/foo", b(''))
        self.fs.setcontents("p/1/2/3/bar", b(''))
        self.fs.makedir("q")
        # Test listing just files, just dirs, and wildcards
        dirs_only = self.fs.listdir(dirs_only=True)
        files_only = self.fs.listdir(files_only=True)
        contains_a = self.fs.listdir(wildcard="*a*")
        self.assertEqual(sorted(dirs_only), [u"p", u"q"])
        self.assertEqual(sorted(files_only), [u"a", u"b", u"bar", u"foo"])
        self.assertEqual(sorted(contains_a), [u"a", u"bar"])
        check_unicode(dirs_only)
        check_unicode(files_only)
        check_unicode(contains_a)
        # Test listing a subdirectory
        d3 = self.fs.listdir("p/1/2/3")
        self.assertEqual(len(d3), 4)
        self.assertEqual(sorted(d3), [u"a", u"b", u"bar", u"foo"])
        check_unicode(d3)
        # Test listing a subdirectory with absoliute and full paths
        d4 = self.fs.listdir("p/1/2/3", absolute=True)
        self.assertEqual(len(d4), 4)
        self.assertEqual(sorted(d4), [u"/p/1/2/3/a", u"/p/1/2/3/b", u"/p/1/2/3/bar", u"/p/1/2/3/foo"])
        check_unicode(d4)
        d4 = self.fs.listdir("p/1/2/3", full=True)
        self.assertEqual(len(d4), 4)
        self.assertEqual(sorted(d4), [u"p/1/2/3/a", u"p/1/2/3/b", u"p/1/2/3/bar", u"p/1/2/3/foo"])
        check_unicode(d4)
        # Test that appropriate errors are raised
        self.assertRaises(ResourceNotFoundError, self.fs.listdir, "zebra")
        self.assertRaises(ResourceInvalidError, self.fs.listdir, "foo")

    def test_listdirinfo(self):
        def check_unicode(items):
            for (nm, info) in items:
                self.assertTrue(isinstance(nm, unicode))

        def check_equal(items, target):
            names = [nm for (nm, info) in items]
            self.assertEqual(sorted(names), sorted(target))
        self.fs.setcontents(u"a", b(''))
        self.fs.setcontents("b", b(''))
        self.fs.setcontents("foo", b(''))
        self.fs.setcontents("bar", b(''))
        # Test listing of the root directory
        d1 = self.fs.listdirinfo()
        self.assertEqual(len(d1), 4)
        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
        check_unicode(d1)
        d1 = self.fs.listdirinfo("")
        self.assertEqual(len(d1), 4)
        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
        check_unicode(d1)
        d1 = self.fs.listdirinfo("/")
        self.assertEqual(len(d1), 4)
        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
        check_unicode(d1)
        # Test listing absolute paths
        d2 = self.fs.listdirinfo(absolute=True)
        self.assertEqual(len(d2), 4)
        check_equal(d2, [u"/a", u"/b", u"/bar", u"/foo"])
        check_unicode(d2)
        # Create some deeper subdirectories, to make sure their
        # contents are not inadvertantly included
        self.fs.makedir("p/1/2/3", recursive=True)
        self.fs.setcontents("p/1/2/3/a", b(''))
        self.fs.setcontents("p/1/2/3/b", b(''))
        self.fs.setcontents("p/1/2/3/foo", b(''))
        self.fs.setcontents("p/1/2/3/bar", b(''))
        self.fs.makedir("q")
        # Test listing just files, just dirs, and wildcards
        dirs_only = self.fs.listdirinfo(dirs_only=True)
        files_only = self.fs.listdirinfo(files_only=True)
        contains_a = self.fs.listdirinfo(wildcard="*a*")
        check_equal(dirs_only, [u"p", u"q"])
        check_equal(files_only, [u"a", u"b", u"bar", u"foo"])
        check_equal(contains_a, [u"a", u"bar"])
        check_unicode(dirs_only)
        check_unicode(files_only)
        check_unicode(contains_a)
        # Test listing a subdirectory
        d3 = self.fs.listdirinfo("p/1/2/3")
        self.assertEqual(len(d3), 4)
        check_equal(d3, [u"a", u"b", u"bar", u"foo"])
        check_unicode(d3)
        # Test listing a subdirectory with absoliute and full paths
        d4 = self.fs.listdirinfo("p/1/2/3", absolute=True)
        self.assertEqual(len(d4), 4)
        check_equal(d4, [u"/p/1/2/3/a", u"/p/1/2/3/b", u"/p/1/2/3/bar", u"/p/1/2/3/foo"])
        check_unicode(d4)
        d4 = self.fs.listdirinfo("p/1/2/3", full=True)
        self.assertEqual(len(d4), 4)
        check_equal(d4, [u"p/1/2/3/a", u"p/1/2/3/b", u"p/1/2/3/bar", u"p/1/2/3/foo"])
        check_unicode(d4)
        # Test that appropriate errors are raised
        self.assertRaises(ResourceNotFoundError, self.fs.listdirinfo, "zebra")
        self.assertRaises(ResourceInvalidError, self.fs.listdirinfo, "foo")

    def test_walk(self):
        self.fs.setcontents('a.txt', b('hello'))
        self.fs.setcontents('b.txt', b('world'))
        self.fs.makeopendir('foo').setcontents('c', b('123'))
        sorted_walk = sorted([(d, sorted(fs)) for (d, fs) in self.fs.walk()])
        self.assertEquals(sorted_walk,
                          [("/", ["a.txt", "b.txt"]),
                           ("/foo", ["c"])])
        #  When searching breadth-first, shallow entries come first
        found_a = False
        for _, files in self.fs.walk(search="breadth"):
            if "a.txt" in files:
                found_a = True
            if "c" in files:
                break
        assert found_a, "breadth search order was wrong"
        #  When searching depth-first, deep entries come first
        found_c = False
        for _, files in self.fs.walk(search="depth"):
            if "c" in files:
                found_c = True
            if "a.txt" in files:
                break
        assert found_c, "depth search order was wrong: " + \
            str(list(self.fs.walk(search="depth")))

    def test_walk_wildcard(self):
        self.fs.setcontents('a.txt', b('hello'))
        self.fs.setcontents('b.txt', b('world'))
        self.fs.makeopendir('foo').setcontents('c', b('123'))
        self.fs.makeopendir('.svn').setcontents('ignored', b(''))
        for dir_path, paths in self.fs.walk(wildcard='*.txt'):
            for path in paths:
                self.assert_(path.endswith('.txt'))
        for dir_path, paths in self.fs.walk(wildcard=lambda fn: fn.endswith('.txt')):
            for path in paths:
                self.assert_(path.endswith('.txt'))

    def test_walk_dir_wildcard(self):
        self.fs.setcontents('a.txt', b('hello'))
        self.fs.setcontents('b.txt', b('world'))
        self.fs.makeopendir('foo').setcontents('c', b('123'))
        self.fs.makeopendir('.svn').setcontents('ignored', b(''))
        for dir_path, paths in self.fs.walk(dir_wildcard=lambda fn: not fn.endswith('.svn')):
            for path in paths:
                self.assert_('.svn' not in path)

    def test_walkfiles(self):
        self.fs.makeopendir('bar').setcontents('a.txt', b('123'))
        self.fs.makeopendir('foo').setcontents('b', b('123'))
        self.assertEquals(sorted(
            self.fs.walkfiles()), ["/bar/a.txt", "/foo/b"])
        self.assertEquals(sorted(self.fs.walkfiles(
            dir_wildcard="*foo*")), ["/foo/b"])
        self.assertEquals(sorted(self.fs.walkfiles(
            wildcard="*.txt")), ["/bar/a.txt"])

    def test_walkdirs(self):
        self.fs.makeopendir('bar').setcontents('a.txt', b('123'))
        self.fs.makeopendir('foo').makeopendir(
            "baz").setcontents('b', b('123'))
        self.assertEquals(sorted(self.fs.walkdirs()), [
                          "/", "/bar", "/foo", "/foo/baz"])
        self.assertEquals(sorted(self.fs.walkdirs(
            wildcard="*foo*")), ["/", "/foo", "/foo/baz"])

    def test_unicode(self):
        alpha = u"\N{GREEK SMALL LETTER ALPHA}"
        beta = u"\N{GREEK SMALL LETTER BETA}"
        self.fs.makedir(alpha)
        self.fs.setcontents(alpha + "/a", b(''))
        self.fs.setcontents(alpha + "/" + beta, b(''))
        self.assertTrue(self.check(alpha))
        self.assertEquals(sorted(self.fs.listdir(alpha)), ["a", beta])

    def test_makedir(self):
        check = self.check
        self.fs.makedir("a")
        self.assertTrue(check("a"))
        self.assertRaises(
            ParentDirectoryMissingError, self.fs.makedir, "a/b/c")
        self.fs.makedir("a/b/c", recursive=True)
        self.assert_(check("a/b/c"))
        self.fs.makedir("foo/bar/baz", recursive=True)
        self.assert_(check("foo/bar/baz"))
        self.fs.makedir("a/b/child")
        self.assert_(check("a/b/child"))
        self.assertRaises(DestinationExistsError, self.fs.makedir, "/a/b")
        self.fs.makedir("/a/b", allow_recreate=True)
        self.fs.setcontents("/a/file", b(''))
        self.assertRaises(ResourceInvalidError, self.fs.makedir, "a/file")

    def test_remove(self):
        self.fs.setcontents("a.txt", b(''))
        self.assertTrue(self.check("a.txt"))
        self.fs.remove("a.txt")
        self.assertFalse(self.check("a.txt"))
        self.assertRaises(ResourceNotFoundError, self.fs.remove, "a.txt")
        self.fs.makedir("dir1")
        self.assertRaises(ResourceInvalidError, self.fs.remove, "dir1")
        self.fs.setcontents("/dir1/a.txt", b(''))
        self.assertTrue(self.check("dir1/a.txt"))
        self.fs.remove("dir1/a.txt")
        self.assertFalse(self.check("/dir1/a.txt"))

    def test_removedir(self):
        check = self.check
        self.fs.makedir("a")
        self.assert_(check("a"))
        self.fs.removedir("a")
        self.assertRaises(ResourceNotFoundError, self.fs.removedir, "a")
        self.assert_(not check("a"))
        self.fs.makedir("a/b/c/d", recursive=True)
        self.assertRaises(DirectoryNotEmptyError, self.fs.removedir, "a/b")
        self.fs.removedir("a/b/c/d")
        self.assert_(not check("a/b/c/d"))
        self.fs.removedir("a/b/c")
        self.assert_(not check("a/b/c"))
        self.fs.removedir("a/b")
        self.assert_(not check("a/b"))
        #  Test recursive removal of empty parent dirs
        self.fs.makedir("foo/bar/baz", recursive=True)
        self.fs.removedir("foo/bar/baz", recursive=True)
        self.assert_(not check("foo/bar/baz"))
        self.assert_(not check("foo/bar"))
        self.assert_(not check("foo"))
        self.fs.makedir("foo/bar/baz", recursive=True)
        self.fs.setcontents("foo/file.txt", b("please don't delete me"))
        self.fs.removedir("foo/bar/baz", recursive=True)
        self.assert_(not check("foo/bar/baz"))
        self.assert_(not check("foo/bar"))
        self.assert_(check("foo/file.txt"))
        #  Ensure that force=True works as expected
        self.fs.makedir("frollic/waggle", recursive=True)
        self.fs.setcontents("frollic/waddle.txt", b("waddlewaddlewaddle"))
        self.assertRaises(DirectoryNotEmptyError, self.fs.removedir, "frollic")
        self.assertRaises(
            ResourceInvalidError, self.fs.removedir, "frollic/waddle.txt")
        self.fs.removedir("frollic", force=True)
        self.assert_(not check("frollic"))
        #  Test removing unicode dirs
        kappa = u"\N{GREEK CAPITAL LETTER KAPPA}"
        self.fs.makedir(kappa)
        self.assert_(self.fs.isdir(kappa))
        self.fs.removedir(kappa)
        self.assertRaises(ResourceNotFoundError, self.fs.removedir, kappa)
        self.assert_(not self.fs.isdir(kappa))
        self.fs.makedir(pathjoin("test", kappa), recursive=True)
        self.assert_(check(pathjoin("test", kappa)))
        self.fs.removedir("test", force=True)
        self.assert_(not check("test"))

    def test_rename(self):
        check = self.check
        # test renaming a file in the same directory
        self.fs.setcontents("foo.txt", b("Hello, World!"))
        self.assert_(check("foo.txt"))
        self.fs.rename("foo.txt", "bar.txt")
        self.assert_(check("bar.txt"))
        self.assert_(not check("foo.txt"))
        # test renaming a directory in the same directory
        self.fs.makedir("dir_a")
        self.fs.setcontents("dir_a/test.txt", b("testerific"))
        self.assert_(check("dir_a"))
        self.fs.rename("dir_a", "dir_b")
        self.assert_(check("dir_b"))
        self.assert_(check("dir_b/test.txt"))
        self.assert_(not check("dir_a/test.txt"))
        self.assert_(not check("dir_a"))
        # test renaming a file into a different directory
        self.fs.makedir("dir_a")
        self.fs.rename("dir_b/test.txt", "dir_a/test.txt")
        self.assert_(not check("dir_b/test.txt"))
        self.assert_(check("dir_a/test.txt"))
        # test renaming a file into a non-existent  directory
        self.assertRaises(ParentDirectoryMissingError,
                          self.fs.rename, "dir_a/test.txt", "nonexistent/test.txt")

    def test_info(self):
        test_str = b("Hello, World!")
        self.fs.setcontents("info.txt", test_str)
        info = self.fs.getinfo("info.txt")
        self.assertEqual(info['size'], len(test_str))
        self.fs.desc("info.txt")
        self.assertRaises(ResourceNotFoundError, self.fs.getinfo, "notafile")
        self.assertRaises(
            ResourceNotFoundError, self.fs.getinfo, "info.txt/inval")

    def test_infokeys(self):
        test_str = b("Hello, World!")
        self.fs.setcontents("info.txt", test_str)
        info = self.fs.getinfo("info.txt")
        for k, v in info.iteritems():
            self.assertEqual(self.fs.getinfokeys('info.txt', k), {k: v})

        test_info = {}
        if 'modified_time' in info:
            test_info['modified_time'] = info['modified_time']
        if 'size' in info:
            test_info['size'] = info['size']
        self.assertEqual(self.fs.getinfokeys('info.txt', 'size', 'modified_time'), test_info)
        self.assertEqual(self.fs.getinfokeys('info.txt', 'thiscantpossiblyexistininfo'), {})

    def test_getsize(self):
        test_str = b("*") * 23
        self.fs.setcontents("info.txt", test_str)
        size = self.fs.getsize("info.txt")
        self.assertEqual(size, len(test_str))

    def test_movefile(self):
        check = self.check
        contents = b(
            "If the implementation is hard to explain, it's a bad idea.")

        def makefile(path):
            self.fs.setcontents(path, contents)

        def checkcontents(path):
            check_contents = self.fs.getcontents(path, "rb")
            self.assertEqual(check_contents, contents)
            return contents == check_contents

        self.fs.makedir("foo/bar", recursive=True)
        makefile("foo/bar/a.txt")
        self.assert_(check("foo/bar/a.txt"))
        self.assert_(checkcontents("foo/bar/a.txt"))
        self.fs.move("foo/bar/a.txt", "foo/b.txt")
        self.assert_(not check("foo/bar/a.txt"))
        self.assert_(check("foo/b.txt"))
        self.assert_(checkcontents("foo/b.txt"))

        self.fs.move("foo/b.txt", "c.txt")
        self.assert_(not check("foo/b.txt"))
        self.assert_(check("/c.txt"))
        self.assert_(checkcontents("/c.txt"))

        makefile("foo/bar/a.txt")
        self.assertRaises(
            DestinationExistsError, self.fs.move, "foo/bar/a.txt", "/c.txt")
        self.assert_(check("foo/bar/a.txt"))
        self.assert_(check("/c.txt"))
        self.fs.move("foo/bar/a.txt", "/c.txt", overwrite=True)
        self.assert_(not check("foo/bar/a.txt"))
        self.assert_(check("/c.txt"))

    def test_movedir(self):
        check = self.check
        contents = b(
            "If the implementation is hard to explain, it's a bad idea.")

        def makefile(path):
            self.fs.setcontents(path, contents)

        self.assertRaises(ResourceNotFoundError, self.fs.movedir, "a", "b")
        self.fs.makedir("a")
        self.fs.makedir("b")
        makefile("a/1.txt")
        makefile("a/2.txt")
        makefile("a/3.txt")
        self.fs.makedir("a/foo/bar", recursive=True)
        makefile("a/foo/bar/baz.txt")

        self.fs.movedir("a", "copy of a")

        self.assert_(self.fs.isdir("copy of a"))
        self.assert_(check("copy of a/1.txt"))
        self.assert_(check("copy of a/2.txt"))
        self.assert_(check("copy of a/3.txt"))
        self.assert_(check("copy of a/foo/bar/baz.txt"))

        self.assert_(not check("a/1.txt"))
        self.assert_(not check("a/2.txt"))
        self.assert_(not check("a/3.txt"))
        self.assert_(not check("a/foo/bar/baz.txt"))
        self.assert_(not check("a/foo/bar"))
        self.assert_(not check("a/foo"))
        self.assert_(not check("a"))

        self.fs.makedir("a")
        self.assertRaises(
            DestinationExistsError, self.fs.movedir, "copy of a", "a")
        self.fs.movedir("copy of a", "a", overwrite=True)
        self.assert_(not check("copy of a"))
        self.assert_(check("a/1.txt"))
        self.assert_(check("a/2.txt"))
        self.assert_(check("a/3.txt"))
        self.assert_(check("a/foo/bar/baz.txt"))

    def test_cant_copy_from_os(self):
        sys_executable = os.path.abspath(os.path.realpath(sys.executable))
        self.assertRaises(FSError, self.fs.copy, sys_executable, "py.exe")

    def test_copyfile(self):
        check = self.check
        contents = b(
            "If the implementation is hard to explain, it's a bad idea.")

        def makefile(path, contents=contents):
            self.fs.setcontents(path, contents)

        def checkcontents(path, contents=contents):
            check_contents = self.fs.getcontents(path, "rb")
            self.assertEqual(check_contents, contents)
            return contents == check_contents

        self.fs.makedir("foo/bar", recursive=True)
        makefile("foo/bar/a.txt")
        self.assert_(check("foo/bar/a.txt"))
        self.assert_(checkcontents("foo/bar/a.txt"))
        # import rpdb2; rpdb2.start_embedded_debugger('password');
        self.fs.copy("foo/bar/a.txt", "foo/b.txt")
        self.assert_(check("foo/bar/a.txt"))
        self.assert_(check("foo/b.txt"))
        self.assert_(checkcontents("foo/bar/a.txt"))
        self.assert_(checkcontents("foo/b.txt"))

        self.fs.copy("foo/b.txt", "c.txt")
        self.assert_(check("foo/b.txt"))
        self.assert_(check("/c.txt"))
        self.assert_(checkcontents("/c.txt"))

        makefile("foo/bar/a.txt", b("different contents"))
        self.assert_(checkcontents("foo/bar/a.txt", b("different contents")))
        self.assertRaises(
            DestinationExistsError, self.fs.copy, "foo/bar/a.txt", "/c.txt")
        self.assert_(checkcontents("/c.txt"))
        self.fs.copy("foo/bar/a.txt", "/c.txt", overwrite=True)
        self.assert_(checkcontents("foo/bar/a.txt", b("different contents")))
        self.assert_(checkcontents("/c.txt", b("different contents")))

    def test_copydir(self):
        check = self.check
        contents = b(
            "If the implementation is hard to explain, it's a bad idea.")

        def makefile(path):
            self.fs.setcontents(path, contents)

        def checkcontents(path):
            check_contents = self.fs.getcontents(path)
            self.assertEqual(check_contents, contents)
            return contents == check_contents

        self.fs.makedir("a")
        self.fs.makedir("b")
        makefile("a/1.txt")
        makefile("a/2.txt")
        makefile("a/3.txt")
        self.fs.makedir("a/foo/bar", recursive=True)
        makefile("a/foo/bar/baz.txt")

        self.fs.copydir("a", "copy of a")
        self.assert_(check("copy of a/1.txt"))
        self.assert_(check("copy of a/2.txt"))
        self.assert_(check("copy of a/3.txt"))
        self.assert_(check("copy of a/foo/bar/baz.txt"))
        checkcontents("copy of a/1.txt")

        self.assert_(check("a/1.txt"))
        self.assert_(check("a/2.txt"))
        self.assert_(check("a/3.txt"))
        self.assert_(check("a/foo/bar/baz.txt"))
        checkcontents("a/1.txt")

        self.assertRaises(DestinationExistsError, self.fs.copydir, "a", "b")
        self.fs.copydir("a", "b", overwrite=True)
        self.assert_(check("b/1.txt"))
        self.assert_(check("b/2.txt"))
        self.assert_(check("b/3.txt"))
        self.assert_(check("b/foo/bar/baz.txt"))
        checkcontents("b/1.txt")

    def test_copydir_with_dotfile(self):
        check = self.check
        contents = b(
            "If the implementation is hard to explain, it's a bad idea.")

        def makefile(path):
            self.fs.setcontents(path, contents)

        self.fs.makedir("a")
        makefile("a/1.txt")
        makefile("a/2.txt")
        makefile("a/.hidden.txt")

        self.fs.copydir("a", "copy of a")
        self.assert_(check("copy of a/1.txt"))
        self.assert_(check("copy of a/2.txt"))
        self.assert_(check("copy of a/.hidden.txt"))

        self.assert_(check("a/1.txt"))
        self.assert_(check("a/2.txt"))
        self.assert_(check("a/.hidden.txt"))

    def test_readwriteappendseek(self):
        def checkcontents(path, check_contents):
            read_contents = self.fs.getcontents(path, "rb")
            self.assertEqual(read_contents, check_contents)
            return read_contents == check_contents
        test_strings = [b("Beautiful is better than ugly."),
                        b("Explicit is better than implicit."),
                        b("Simple is better than complex.")]
        all_strings = b("").join(test_strings)

        self.assertRaises(ResourceNotFoundError, self.fs.open, "a.txt", "r")
        self.assert_(not self.fs.exists("a.txt"))
        f1 = self.fs.open("a.txt", "wb")
        pos = 0
        for s in test_strings:
            f1.write(s)
            pos += len(s)
            self.assertEqual(pos, f1.tell())
        f1.close()
        self.assert_(self.fs.exists("a.txt"))
        self.assert_(checkcontents("a.txt", all_strings))

        f2 = self.fs.open("b.txt", "wb")
        f2.write(test_strings[0])
        f2.close()
        self.assert_(checkcontents("b.txt", test_strings[0]))
        f3 = self.fs.open("b.txt", "ab")
        # On win32, tell() gives zero until you actually write to the file
        # self.assertEquals(f3.tell(),len(test_strings[0]))
        f3.write(test_strings[1])
        self.assertEquals(f3.tell(), len(test_strings[0])+len(test_strings[1]))
        f3.write(test_strings[2])
        self.assertEquals(f3.tell(), len(all_strings))
        f3.close()
        self.assert_(checkcontents("b.txt", all_strings))
        f4 = self.fs.open("b.txt", "wb")
        f4.write(test_strings[2])
        f4.close()
        self.assert_(checkcontents("b.txt", test_strings[2]))
        f5 = self.fs.open("c.txt", "wb")
        for s in test_strings:
            f5.write(s+b("\n"))
        f5.close()
        f6 = self.fs.open("c.txt", "rb")
        for s, t in zip(f6, test_strings):
            self.assertEqual(s, t+b("\n"))
        f6.close()
        f7 = self.fs.open("c.txt", "rb")
        f7.seek(13)
        word = f7.read(6)
        self.assertEqual(word, b("better"))
        f7.seek(1, os.SEEK_CUR)
        word = f7.read(4)
        self.assertEqual(word, b("than"))
        f7.seek(-9, os.SEEK_END)
        word = f7.read(7)
        self.assertEqual(word, b("complex"))
        f7.close()
        self.assertEqual(self.fs.getcontents("a.txt", "rb"), all_strings)

    def test_truncate(self):
        def checkcontents(path, check_contents):
            read_contents = self.fs.getcontents(path, "rb")
            self.assertEqual(read_contents, check_contents)
            return read_contents == check_contents
        self.fs.setcontents("hello", b("world"))
        checkcontents("hello", b("world"))
        self.fs.setcontents("hello", b("hi"))
        checkcontents("hello", b("hi"))
        self.fs.setcontents("hello", b("1234567890"))
        checkcontents("hello", b("1234567890"))
        with self.fs.open("hello", "rb+") as f:
            f.truncate(7)
        checkcontents("hello", b("1234567"))
        with self.fs.open("hello", "rb+") as f:
            f.seek(5)
            f.truncate()
        checkcontents("hello", b("12345"))

    def test_truncate_to_larger_size(self):
        with self.fs.open("hello", "wb") as f:
            f.truncate(30)

        self.assertEquals(self.fs.getsize("hello"), 30)

        # Some file systems (FTPFS) don't support both reading and writing
        if self.fs.getmeta('file.read_and_write', True):
            with self.fs.open("hello", "rb+") as f:
                f.seek(25)
                f.write(b("123456"))

            with self.fs.open("hello", "rb") as f:
                f.seek(25)
                self.assertEquals(f.read(), b("123456"))

    def test_write_past_end_of_file(self):
        if self.fs.getmeta('file.read_and_write', True):
            with self.fs.open("write_at_end", "wb") as f:
                f.seek(25)
                f.write(b("EOF"))
            with self.fs.open("write_at_end", "rb") as f:
                self.assertEquals(f.read(), b("\x00")*25 + b("EOF"))

    def test_with_statement(self):
        #  This is a little tricky since 'with' is actually new syntax.
        #  We use eval() to make this method safe for old python versions.
        import sys
        if sys.version_info[0] >= 2 and sys.version_info[1] >= 5:
            #  A successful 'with' statement
            contents = "testing the with statement"
            code = "from __future__ import with_statement\n"
            code += "with self.fs.open('f.txt','wb-') as testfile:\n"
            code += "    testfile.write(contents)\n"
            code += "self.assertEquals(self.fs.getcontents('f.txt', 'rb'),contents)"
            code = compile(code, "<string>", 'exec')
            eval(code)
            # A 'with' statement raising an error
            contents = "testing the with statement"
            code = "from __future__ import with_statement\n"
            code += "with self.fs.open('f.txt','wb-') as testfile:\n"
            code += "    testfile.write(contents)\n"
            code += "    raise ValueError\n"
            code = compile(code, "<string>", 'exec')
            self.assertRaises(ValueError, eval, code, globals(), locals())
            self.assertEquals(self.fs.getcontents('f.txt', 'rb'), contents)

    def test_pickling(self):
        if self.fs.getmeta('pickle_contents', True):
            self.fs.setcontents("test1", b("hello world"))
            fs2 = pickle.loads(pickle.dumps(self.fs))
            self.assert_(fs2.isfile("test1"))
            fs3 = pickle.loads(pickle.dumps(self.fs, -1))
            self.assert_(fs3.isfile("test1"))
        else:
            # Just make sure it doesn't throw an exception
            fs2 = pickle.loads(pickle.dumps(self.fs))

    def test_big_file(self):
        """Test handling of a big file (1MB)"""
        chunk_size = 1024 * 256
        num_chunks = 4

        def chunk_stream():
            """Generate predictable-but-randomy binary content."""
            r = random.Random(0)
            randint = r.randint
            int2byte = six.int2byte
            for _i in xrange(num_chunks):
                c = b("").join(int2byte(randint(
                    0, 255)) for _j in xrange(chunk_size//8))
                yield c * 8
                f = self.fs.open("bigfile", "wb")
                try:
                    for chunk in chunk_stream():
                        f.write(chunk)
                finally:
                    f.close()
                chunks = chunk_stream()
                f = self.fs.open("bigfile", "rb")
                try:
                    try:
                        while True:
                            if chunks.next() != f.read(chunk_size):
                                assert False, "bigfile was corrupted"
                    except StopIteration:
                        if f.read() != b(""):
                            assert False, "bigfile was corrupted"
                finally:
                    f.close()

    def test_settimes(self):
        def cmp_datetimes(d1, d2):
            """Test datetime objects are the same to within the timestamp accuracy"""
            dts1 = time.mktime(d1.timetuple())
            dts2 = time.mktime(d2.timetuple())
            return int(dts1) == int(dts2)
        d1 = datetime.datetime(2010, 6, 20, 11, 0, 9, 987699)
        d2 = datetime.datetime(2010, 7, 5, 11, 0, 9, 500000)
        self.fs.setcontents('/dates.txt', b('check dates'))
        # If the implementation supports settimes, check that the times
        # can be set and then retrieved
        try:
            self.fs.settimes('/dates.txt', d1, d2)
        except UnsupportedError:
            pass
        else:
            info = self.fs.getinfo('/dates.txt')
            self.assertTrue(cmp_datetimes(d1, info['accessed_time']))
            self.assertTrue(cmp_datetimes(d2, info['modified_time']))

    def test_removeroot(self):
        self.assertRaises(RemoveRootError, self.fs.removedir, "/")

    def test_zero_read(self):
        """Test read(0) returns empty string"""
        self.fs.setcontents('foo.txt', b('Hello, World'))
        with self.fs.open('foo.txt', 'rb') as f:
            self.assert_(len(f.read(0)) == 0)
        with self.fs.open('foo.txt', 'rt') as f:
            self.assert_(len(f.read(0)) == 0)

# May be disabled - see end of file


class ThreadingTestCases(object):
    """Testcases for thread-safety of FS implementations."""

    #  These are either too slow to be worth repeating,
    #  or cannot possibly break cross-thread.
    _dont_retest = ("test_pickling", "test_multiple_overwrite",)

    __lock = threading.RLock()

    def _yield(self):
        # time.sleep(0.001)
        # Yields without a delay
        time.sleep(0)

    def _lock(self):
        self.__lock.acquire()

    def _unlock(self):
        self.__lock.release()

    def _makeThread(self, func, errors):
        def runThread():
            try:
                func()
            except Exception:
                errors.append(sys.exc_info())
        thread = threading.Thread(target=runThread)
        thread.daemon = True
        return thread

    def _runThreads(self, *funcs):
        check_interval = sys.getcheckinterval()
        sys.setcheckinterval(1)
        try:
            errors = []
            threads = [self._makeThread(f, errors) for f in funcs]
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            for (c, e, t) in errors:
                raise e, None, t
        finally:
            sys.setcheckinterval(check_interval)

    def test_setcontents_threaded(self):
        def setcontents(name, contents):
            f = self.fs.open(name, "wb")
            self._yield()
            try:
                f.write(contents)
                self._yield()
            finally:
                f.close()

        def thread1():
            c = b("thread1 was 'ere")
            setcontents("thread1.txt", c)
            self.assertEquals(self.fs.getcontents("thread1.txt", 'rb'), c)

        def thread2():
            c = b("thread2 was 'ere")
            setcontents("thread2.txt", c)
            self.assertEquals(self.fs.getcontents("thread2.txt", 'rb'), c)
        self._runThreads(thread1, thread2)

    def test_setcontents_threaded_samefile(self):
        def setcontents(name, contents):
            f = self.fs.open(name, "wb")
            self._yield()
            try:
                f.write(contents)
                self._yield()
            finally:
                f.close()

        def thread1():
            c = b("thread1 was 'ere")
            setcontents("threads.txt", c)
            self._yield()
            self.assertEquals(self.fs.listdir("/"), ["threads.txt"])

        def thread2():
            c = b("thread2 was 'ere")
            setcontents("threads.txt", c)
            self._yield()
            self.assertEquals(self.fs.listdir("/"), ["threads.txt"])

        def thread3():
            c = b("thread3 was 'ere")
            setcontents("threads.txt", c)
            self._yield()
            self.assertEquals(self.fs.listdir("/"), ["threads.txt"])
        try:
            self._runThreads(thread1, thread2, thread3)
        except ResourceLockedError:
            # that's ok, some implementations don't support concurrent writes
            pass

    def test_cases_in_separate_dirs(self):
        class TestCases_in_subdir(self.__class__, unittest.TestCase):
            """Run all testcases against a subdir of self.fs"""
            def __init__(this, subdir):
                super(TestCases_in_subdir, this).__init__("test_listdir")
                this.subdir = subdir
                for meth in dir(this):
                    if not meth.startswith("test_"):
                        continue
                    if meth in self._dont_retest:
                        continue
                    if not hasattr(FSTestCases, meth):
                        continue
                    if self.fs.exists(subdir):
                        self.fs.removedir(subdir, force=True)
                    self.assertFalse(self.fs.isdir(subdir))
                    self.assertTrue(self.fs.isdir("/"))
                    self.fs.makedir(subdir)
                    self._yield()
                    getattr(this, meth)()

            @property
            def fs(this):
                return self.fs.opendir(this.subdir)

            def check(this, p):
                return self.check(pathjoin(this.subdir, relpath(p)))

        def thread1():
            TestCases_in_subdir("thread1")

        def thread2():
            TestCases_in_subdir("thread2")

        def thread3():
            TestCases_in_subdir("thread3")
        self._runThreads(thread1, thread2, thread3)

    def test_makedir_winner(self):
        errors = []

        def makedir():
            try:
                self.fs.makedir("testdir")
            except DestinationExistsError, e:
                errors.append(e)

        def makedir_noerror():
            try:
                self.fs.makedir("testdir", allow_recreate=True)
            except DestinationExistsError, e:
                errors.append(e)

        def removedir():
            try:
                self.fs.removedir("testdir")
            except (ResourceNotFoundError, ResourceLockedError), e:
                errors.append(e)
        # One thread should succeed, one should error
        self._runThreads(makedir, makedir)
        self.assertEquals(len(errors), 1)
        self.fs.removedir("testdir")
        # One thread should succeed, two should error
        errors = []
        self._runThreads(makedir, makedir, makedir)
        if len(errors) != 2:
            raise AssertionError(errors)
        self.fs.removedir("testdir")
        # All threads should succeed
        errors = []
        self._runThreads(makedir_noerror, makedir_noerror, makedir_noerror)
        self.assertEquals(len(errors), 0)
        self.assertTrue(self.fs.isdir("testdir"))
        self.fs.removedir("testdir")
        # makedir() can beat removedir() and vice-versa
        errors = []
        self._runThreads(makedir, removedir)
        if self.fs.isdir("testdir"):
            self.assertEquals(len(errors), 1)
            self.assertFalse(isinstance(errors[0], DestinationExistsError))
            self.fs.removedir("testdir")
        else:
            self.assertEquals(len(errors), 0)

    def test_concurrent_copydir(self):
        self.fs.makedir("a")
        self.fs.makedir("a/b")
        self.fs.setcontents("a/hello.txt", b("hello world"))
        self.fs.setcontents("a/guido.txt", b("is a space alien"))
        self.fs.setcontents("a/b/parrot.txt", b("pining for the fiords"))

        def copydir():
            self._yield()
            self.fs.copydir("a", "copy of a")

        def copydir_overwrite():
            self._yield()
            self.fs.copydir("a", "copy of a", overwrite=True)
        # This should error out since we're not overwriting
        self.assertRaises(
            DestinationExistsError, self._runThreads, copydir, copydir)
        self.assert_(self.fs.isdir('a'))
        self.assert_(self.fs.isdir('a'))
        copydir_overwrite()
        self.assert_(self.fs.isdir('a'))
        # This should run to completion and give a valid state, unless
        # files get locked when written to.
        try:
            self._runThreads(copydir_overwrite, copydir_overwrite)
        except ResourceLockedError:
            pass
        self.assertTrue(self.fs.isdir("copy of a"))
        self.assertTrue(self.fs.isdir("copy of a/b"))
        self.assertEqual(self.fs.getcontents(
            "copy of a/b/parrot.txt", 'rb'), b("pining for the fiords"))
        self.assertEqual(self.fs.getcontents(
            "copy of a/hello.txt", 'rb'), b("hello world"))
        self.assertEqual(self.fs.getcontents(
            "copy of a/guido.txt", 'rb'), b("is a space alien"))

    def test_multiple_overwrite(self):
        contents = [b("contents one"), b(
            "contents the second"), b("number three")]

        def thread1():
            for i in xrange(30):
                for c in contents:
                    self.fs.setcontents("thread1.txt", c)
                    self.assertEquals(self.fs.getsize("thread1.txt"), len(c))
                    self.assertEquals(self.fs.getcontents(
                        "thread1.txt", 'rb'), c)

        def thread2():
            for i in xrange(30):
                for c in contents:
                    self.fs.setcontents("thread2.txt", c)
                    self.assertEquals(self.fs.getsize("thread2.txt"), len(c))
                    self.assertEquals(self.fs.getcontents(
                        "thread2.txt", 'rb'), c)
        self._runThreads(thread1, thread2)

# Uncomment to temporarily disable threading tests
# class ThreadingTestCases(object):
#    _dont_retest = ()



from traits.api import Dict, HasTraits

from apptools.selection.errors import (
    ProviderNotRegisteredError, IDConflictError, ListenerNotConnectedError
)


class SelectionService(HasTraits):
    """ The selection service connects selection providers and listeners.

    The selection service is a register of selection providers, i.e., objects
    that publish their current selection.

    Selections can be requested actively, by explicitly requesting the current
    selection in a provider (:meth:`get_selection(id)`), or passively by
    connecting selection listeners.
    """

    #### 'SelectionService' protocol ##########################################

    def add_selection_provider(self, provider):
        """ Add a selection provider.

        The provider is identified by its ID. If a provider with the same
        ID has been already registered, an :class:`~.IDConflictError`
        is raised.

        Arguments:
            provider -- ISelectionProvider
                The selection provider added to the internal registry.

        """
        provider_id = provider.provider_id
        if self.has_selection_provider(provider_id):
            raise IDConflictError(provider_id=provider_id)

        self._providers[provider_id] = provider

        if provider_id in self._listeners:
            self._connect_all_listeners(provider_id)

    def has_selection_provider(self, provider_id):
        """ Has a provider with the given ID been registered? """
        return provider_id in self._providers

    def remove_selection_provider(self, provider):
        """ Remove a selection provider.

        If the provider has not been registered, a
        :class:`~.ProviderNotRegisteredError` is raised.

        Arguments:
            provider -- ISelectionProvider
                The selection provider added to the internal registry.
        """
        provider_id = provider.provider_id
        self._raise_if_not_registered(provider_id)

        if provider_id in self._listeners:
            self._disconnect_all_listeners(provider_id)

        del self._providers[provider_id]

    def get_selection(self, provider_id):
        """ Return the current selection of the provider with the given ID.

        If a provider with that ID has not been registered, a
        :class:`~.ProviderNotRegisteredError` is raised.

        Arguments:
            provider_id -- str
                The selection provider ID.

        Returns:
            selection -- ISelection
                The current selection of the provider.
        """
        self._raise_if_not_registered(provider_id)
        provider = self._providers[provider_id]
        return provider.get_selection()

    def set_selection(self, provider_id, items, ignore_missing=False):
        """ Set the current selection in a provider to the given items.

        If a provider with the given ID has not been registered, a
        :class:`~.ProviderNotRegisteredError` is raised.

        If ``ignore_missing`` is ``True``, items that are not available in the
        selection provider are silently ignored. If it is ``False`` (default),
        a :class:`ValueError` should be raised.

        Arguments:
            provider_id -- str
                The selection provider ID.

            items -- list
                List of items to be selected.

            ignore_missing -- bool
                If ``False`` (default), the provider raises an exception if any
                of the items in ``items`` is not available to be selected.
                Otherwise, missing elements are silently ignored, and the rest
                is selected.
        """
        self._raise_if_not_registered(provider_id)
        provider = self._providers[provider_id]
        return provider.set_selection(items, ignore_missing=ignore_missing)

    def connect_selection_listener(self, provider_id, func):
        """ Connect a listener to selection events from a specific provider.

        The signature if the listener callback is ``func(i_selection)``.
        The listener is called:

        1) When a provider with the given ID is registered, with its initial
           selection as argument, or

        2) whenever the provider fires a selection event.

        It is perfectly valid to connect a listener before a provider with the
        given ID is registered. The listener will remain connected even if
        the provider is repeatedly connected and disconnected.

        Arguments:
            provider_id -- str
                The selection provider ID.
            func -- callable(i_selection)
                A callable object that is notified when the selection changes.
        """
        self._listeners.setdefault(provider_id, [])
        self._listeners[provider_id].append(func)

        if self.has_selection_provider(provider_id):
            self._toggle_listener(provider_id, func, remove=False)

    def disconnect_selection_listener(self, provider_id, func):
        """ Disconnect a listener from a specific provider.

        Arguments:
            provider_id -- str
                The selection provider ID.
            func -- callable(provider_id, i_selection)
                A callable object that is notified when the selection changes.
        """

        if self.has_selection_provider(provider_id):
            self._toggle_listener(provider_id, func, remove=True)

        try:
            self._listeners[provider_id].remove(func)
        except (ValueError, KeyError):
            raise ListenerNotConnectedError(provider_id=provider_id,
                                            listener=func)

    #### Private protocol #####################################################

    _listeners = Dict()

    _providers = Dict()

    def _toggle_listener(self, provider_id, func, remove):
        provider = self._providers[provider_id]
        provider.on_trait_change(func, 'selection', remove=remove)

    def _connect_all_listeners(self, provider_id):
        """ Connect all listeners connected to a provider.

        As soon as they are connected, they receive the initial selection.
        """
        provider = self._providers[provider_id]
        selection = provider.get_selection()
        for func in self._listeners[provider_id]:
            self._toggle_listener(provider_id, func, remove=False)
            # FIXME: make this robust to notifications that raise exceptions.
            # Can we send the error to the traits exception hook?
            func(selection)

    def _disconnect_all_listeners(self, provider_id):
        for func in self._listeners[provider_id]:
            self._toggle_listener(provider_id, func, remove=True)

    def _raise_if_not_registered(self, provider_id):
        if not self.has_selection_provider(provider_id):
            raise ProviderNotRegisteredError(provider_id=provider_id)



#-------------------------------------------------------------------------------
#
#  A simple scatter-plot template defined as a test for the template package.
#
#  Written by: David C. Morrill
#  (based on the original cp.plot geo_scatter_plot.py file)
#
#  Date: 07/30/2007
#
#  (c) Copy 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A simple scatter-plot template defined as a test for the template package.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Undefined

from traitsui.api \
    import View, VGroup, Item, Label, Theme, TextEditor

from traitsui.wx.themed_slider_editor \
    import ThemedSliderEditor

from traitsui.wx.themed_text_editor \
    import ThemedTextEditor

from chaco.api \
    import ScatterPlot, ArrayPlotData, Plot

from chaco.tools.api \
    import PanTool, SimpleZoom

from enable.api \
    import ColorTrait

from chaco.scatter_markers \
    import marker_trait

from apptools.template.api \
    import Template, TRange, TStr, TDerived, TDataSource

from apptools.template.impl.api \
    import TemplateDataSource, ValueDataNameItem

from enable_editor \
    import EnableEditor

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# Template color trait:
TColor = ColorTrait( template = 'copy' )

#-------------------------------------------------------------------------------
#  'ScatterPlot' class:
#-------------------------------------------------------------------------------

class ScatterPlot ( Template ):

    #-- Template Traits --------------------------------------------------------

    # The plot index data source:
    index = TDataSource

    # The plot value data source:
    value = TDataSource

    # The title of the plot:
    title = TStr( 'Scatter Plot' )

    # The type of marker to use.  This is a mapped trait using strings as the
    # keys:
    marker = marker_trait( template = 'copy', event = 'update' )

    # The pixel size of the marker (doesn't include the thickness of the
    # outline):
    marker_size = TRange( 1, 5, 1, event = 'update' )

    # The thickness, in pixels, of the outline to draw around the marker.  If
    # this is 0, no outline will be drawn.
    line_width = TRange( 0.0, 5.0, 1.0 )

    # The fill color of the marker:
    color = TColor( 'red', event = 'update' )

    # The color of the outline to draw around the marker
    outline_color = TColor( 'black', event = 'update' )

    #-- Derived Traits ---------------------------------------------------------

    plot = TDerived # Instance( ScatterPlot )

    #-- Traits UI Views --------------------------------------------------------

    # The scatter plot view:
    template_view = View(
        VGroup(
            Item( 'title',
                  show_label = False,
                  style      = 'readonly',
                  editor     = ThemedTextEditor(
                                 theme = Theme( '@GBB', alignment = 'center' ) )
            ),
            Item( 'plot',
                  show_label = False,
                  resizable  = True,
                  editor     = EnableEditor(),
                  item_theme = Theme( '@GF5', margins = 0 )
            )
        ),
        resizable = True
    )

    # The scatter plot options view:
    options_view = View(
        VGroup(
            VGroup(
                Label( 'Scatter Plot Options',
                       item_theme = Theme( '@GBB', alignment = 'center' ) ),
                show_labels = False
            ),
            VGroup(
                Item( 'title', editor = TextEditor() ),
                Item( 'marker' ),
                Item( 'marker_size', editor = ThemedSliderEditor() ),
                Item( 'line_width',
                      label  = 'Line Width',
                      editor = ThemedSliderEditor() ),
                Item( 'color',         label = 'Fill Color' ),
                Item( 'outline_color', label = 'Outline Color' ),
                group_theme = Theme( '@GF5', margins = ( -5, -1 ) ),
                item_theme  = Theme( '@G0B', margins = 0 )
            )
        )
    )

    #-- Default Values ---------------------------------------------------------

    def _index_default ( self ):
        """ Returns the default value for the 'index' trait.
        """
        return TemplateDataSource(
                   items       = [ ValueDataNameItem( name    = 'index',
                                                      flatten = True ) ],
                   description = 'Scatter Plot Index' )

    def _value_default ( self ):
        """ Returns the default value for the 'value' trait.
        """
        return TemplateDataSource(
                   items       = [ ValueDataNameItem( name    = 'value',
                                                      flatten = True ) ],
                   description = 'Scatter Plot Value' )

    #-- ITemplate Interface Implementation -------------------------------------

    def activate_template ( self ):
        """ Converts all contained 'TDerived' objects to real objects using the
            template traits of the object. This method must be overridden in
            subclasses.

            Returns
            -------
            None
        """
        # If our data sources are still unbound, then just exit; someone must
        # have marked them as optional:
        if ((self.index.context_data is Undefined) or
            (self.value.context_data is Undefined)):
            return

        # Create a plot data object and give it this data:
        pd = ArrayPlotData()
        pd.set_data( 'index', self.index.context_data )
        pd.set_data( 'value', self.value.context_data )

        # Create the plot:
        self.plot = plot = Plot( pd )
        plot.plot( ( 'index', 'value' ),
                   type           = 'scatter',
                   index_sort     = 'ascending',
                   marker         = self.marker,
                   color          = self.color,
                   outline_color  = self.outline_color,
                   marker_size    = self.marker_size,
                   line_width     = self.line_width,
                   bgcolor        = 'white' )
        plot.set(  padding_left   = 50,
                   padding_right  = 0,
                   padding_top    = 0,
                   padding_bottom = 20 )

        # Attach some tools to the plot:
        plot.tools.append( PanTool( plot, constrain_key = 'shift' ) )
        zoom = SimpleZoom( component = plot, tool_mode = 'box',
                           always_on = False )
        plot.overlays.append( zoom )

    #-- Trait Event Handlers ---------------------------------------------------

    def _update_changed ( self ):
        """ Handles a plot option being changed.
        """
        self.plot = Undefined




#!/usr/bin/env python
from fs.commands.runner import Command
import sys
import platform
import os
import os.path


platform = platform.system()


class FSMount(Command):

    if platform == "Windows":
        usage = """fsmount [OPTIONS]... [FS] [DRIVE LETTER]
or fsmount -u [DRIVER LETTER]
Mounts a filesystem on a drive letter"""
    else:
        usage = """fsmount [OPTIONS]... [FS] [SYSTEM PATH]
or fsmount -u [SYSTEM PATH]
Mounts a file system on a system path"""

    version = "1.0"

    def get_optparse(self):
        optparse = super(FSMount, self).get_optparse()
        optparse.add_option('-f', '--foreground', dest='foreground', action="store_true", default=False,
                            help="run the mount process in the foreground", metavar="FOREGROUND")
        optparse.add_option('-u', '--unmount', dest='unmount', action="store_true", default=False,
                            help="unmount path", metavar="UNMOUNT")
        optparse.add_option('-n', '--nocache', dest='nocache', action="store_true", default=False,
                            help="do not cache network filesystems", metavar="NOCACHE")

        return optparse

    def do_run(self, options, args):

        windows = platform == "Windows"

        if options.unmount:

            if windows:

                try:
                    mount_path = args[0][:1]
                except IndexError:
                    self.error('Driver letter required\n')
                    return 1

                from fs.expose import dokan
                mount_path = mount_path[:1].upper()
                self.output('unmounting %s:...\n' % mount_path, True)
                dokan.unmount(mount_path)
                return

            else:
                try:
                    mount_path = args[0]
                except IndexError:
                    self.error(self.usage + '\n')
                    return 1

                from fs.expose import fuse
                self.output('unmounting %s...\n' % mount_path, True)
                fuse.unmount(mount_path)
                return

        try:
            fs_url = args[0]
        except IndexError:
            self.error(self.usage + '\n')
            return 1

        try:
            mount_path = args[1]
        except IndexError:
            if windows:
                mount_path = mount_path[:1].upper()
                self.error(self.usage + '\n')
            else:
                self.error(self.usage + '\n')
            return 1

        fs, path = self.open_fs(fs_url, create_dir=True)
        if path:
            if not fs.isdir(path):
                self.error('%s is not a directory on %s' % (fs_url, fs))
                return 1
            fs = fs.opendir(path)
            path = '/'
        if not options.nocache:
            fs.cache_hint(True)

        if windows:
            from fs.expose import dokan

            if len(mount_path) > 1:
                self.error('Driver letter should be one character')
                return 1

            self.output("Mounting %s on %s:\n" % (fs, mount_path), True)
            flags = dokan.DOKAN_OPTION_REMOVABLE
            if options.debug:
                flags |= dokan.DOKAN_OPTION_DEBUG | dokan.DOKAN_OPTION_STDERR

            mp = dokan.mount(fs,
                             mount_path,
                             numthreads=5,
                             foreground=options.foreground,
                             flags=flags,
                             volname=str(fs))

        else:
            if not os.path.exists(mount_path):
                try:
                    os.makedirs(mount_path)
                except:
                    pass

            from fs.expose import fuse
            self.output("Mounting %s on %s\n" % (fs, mount_path), True)

            if options.foreground:
                fuse_process = fuse.mount(fs,
                                          mount_path,
                                          foreground=True)
            else:
                if not os.fork():
                    mp = fuse.mount(fs,
                                    mount_path,
                                    foreground=True)
                else:
                    fs.close = lambda: None

def run():
    return FSMount().run()

if __name__ == "__main__":
    sys.exit(run())
